import torch
import torchutil

import emphases


###############################################################################
# Training
###############################################################################


@torchutil.notify('train')
def train(dataset, directory, gpu=None):
    """Train a model"""

    # Get torch device
    device = torch.device('cpu' if gpu is None else f'cuda:{gpu}')

    #######################
    # Create data loaders #
    #######################

    torch.manual_seed(emphases.RANDOM_SEED)

    # Training data
    train_loader = emphases.data.loader(dataset, 'train', gpu)

    # Validation data
    if emphases.VALIDATION_DATASET == 'buckeye':

        # This is just for generating scaling law plots for the paper
        valid_loader = emphases.data.loader('buckeye', 'test', gpu)

    else:

        valid_loader = emphases.data.loader(dataset, 'valid', gpu)

    ################
    # Create model #
    ################

    model = emphases.Model().to(device)

    ####################
    # Create optimizer #
    ####################

    optimizer = torch.optim.Adam(model.parameters())

    ##############################
    # Maybe load from checkpoint #
    ##############################

    path = torchutil.checkpoint.latest_path(directory)

    if path is not None:

        # Load model
        model, optimizer, state = torchutil.checkpoint.load(
            path,
            model,
            optimizer)
        epoch = state['epoch']
        step = state['step']
        score = state['score']
        best = state['best']

    else:

        # Train from scratch
        epoch, step, score, best = 0, 0, 0., 0.

    #########
    # Train #
    #########

    # Automatic mixed precision (amp) gradient scaler
    scaler = torch.cuda.amp.GradScaler()

    # Setup progress bar
    progress = torchutil.iterator(
        range(step, emphases.NUM_STEPS),
        f'Training {emphases.CONFIG}',
        step,
        emphases.NUM_STEPS)
    while step < emphases.NUM_STEPS:

        # Seed sampler
        train_loader.batch_sampler.set_epoch(epoch)

        for batch in train_loader:

            # Unpack batch
            (
                features,
                frame_lengths,
                word_bounds,
                word_lengths,
                targets,
                _,           # alignment
                _,           # audio
                _            # stem
            ) = batch

            # Copy to GPU
            features = features.to(device)
            frame_lengths = frame_lengths.to(device)
            word_bounds = word_bounds.to(device)
            word_lengths = word_lengths.to(device)
            targets = targets.to(device)
            with torch.autocast(device.type):

                # Forward pass
                scores = model(
                    features,
                    frame_lengths,
                    word_bounds,
                    word_lengths)

                # Compute loss
                train_loss = loss(
                    scores,
                    targets,
                    frame_lengths,
                    word_bounds,
                    word_lengths,
                    training=True)

            ##################
            # Optimize model #
            ##################

            optimizer.zero_grad()

            # Backward pass
            scaler.scale(train_loss).backward()

            # Update weights
            scaler.step(optimizer)

            # Update gradient scaler
            scaler.update()

            ############
            # Evaluate #
            ############

            if step % emphases.LOG_INTERVAL == 0:
                score = evaluate(
                    directory,
                    step,
                    model,
                    gpu,
                    'valid',
                    valid_loader)

            ###################
            # Save checkpoint #
            ###################

            if step >= 300 and score > best:
                torchutil.checkpoint.save(
                    directory / f'{step:08d}.pt',
                    model,
                    optimizer,
                    epoch=epoch,
                    step=step,
                    score=score,
                    best=best)
                best = score

            # End training after a certain number of steps
            if step >= emphases.NUM_STEPS:
                break

            # Update training step count
            step += 1

            # Update progress bar
            progress.update()

        # Update epoch count
        epoch += 1

    # Close progress bar
    progress.close()

    # Save final model
    torchutil.checkpoint.save(
        directory / f'{step:08d}.pt',
        model,
        optimizer,
        epoch=epoch,
        step=step,
        score=score,
        best=best)


###############################################################################
# Evaluation
###############################################################################


def evaluate(directory, step, model, gpu, condition, loader):
    """Perform model evaluation"""
    device = 'cpu' if gpu is None else f'cuda:{gpu}'

    # Tensorboard audio and figures
    waveforms, figures = {}, {}

    # Prepare model for inference
    with emphases.inference_context(model):

        # Cache results to evaluate
        results = []
        for i, batch in enumerate(loader):

            # Unpack batch
            (
                features,
                frame_lengths,
                word_bounds,
                word_lengths,
                targets,
                alignments,
                audio,
                stems
            ) = batch

            # Copy to GPU
            features = features.to(device)
            frame_lengths = frame_lengths.to(device)
            word_bounds = word_bounds.to(device)
            word_lengths = word_lengths.to(device)
            targets = targets.to(device)

            # Forward pass
            logits = model(
                features,
                frame_lengths,
                word_bounds,
                word_lengths)

            # Cache results
            results.append((
                logits.detach().cpu(),
                targets.detach().cpu(),
                word_lengths.detach().cpu()))

            # Add audio and figures
            if condition == 'valid' and i < emphases.PLOT_EXAMPLES:

                # Postprocess network output
                scores = emphases.postprocess(logits)

                # Add audio
                samples = emphases.convert.frames_to_samples(frame_lengths[0])
                waveforms[f'audio/{stems[0]}'] = audio[0, :, :samples]

                # Add figure
                figures[stems[0]] = emphases.plot.scores(
                    alignments[0],
                    scores[0, 0, :word_lengths[0]].cpu(),
                    targets[0, 0, :word_lengths[0]].cpu())

            # Stop when we exceed some number of batches
            if i + 1 == emphases.LOG_STEPS:
                break

        # Setup batch statistics
        target_stats = emphases.evaluate.metrics.Statistics()
        predicted_stats = emphases.evaluate.metrics.Statistics()

        # Update statistics
        for logits, targets, word_lengths in results:
            target_stats.update(
                targets.to(device),
                word_lengths.to(device))
            predicted_stats.update(
                emphases.postprocess(logits.to(device)),
                word_lengths.to(device))

        # Setup evaluation metrics
        metrics = emphases.evaluate.Metrics(predicted_stats, target_stats)

        # Update metrics
        for logits, targets, word_lengths in results:
            metrics.update(
                logits.to(device),
                targets.to(device),
                word_lengths.to(device))

    # Format results
    scalars = {
        f'{key}/{condition}': value for key, value in metrics().items()}

    # Write to tensorboard
    torchutil.tensorboard.update(
        directory,
        step,
        scalars=scalars,
        figures=figures,
        audio=waveforms,
        sample_rate=emphases.SAMPLE_RATE)

    # Return Pearson correlation
    return scalars[f'pearson_correlation/{condition}']


###############################################################################
# Loss function
###############################################################################


def loss(
    scores,
    targets,
    frame_lengths,
    word_bounds,
    word_lengths,
    training=False,
    loss_fn=emphases.LOSS):
    """Compute masked loss"""
    if training and emphases.DOWNSAMPLE_LOCATION == 'inference':

        # If we are not downsampling the network output before the loss, we
        # must upsample the targets
        targets = emphases.upsample(
            targets,
            word_bounds,
            word_lengths,
            frame_lengths)

        # Linear interpolation can cause out-of-range
        if emphases.UPSAMPLE_METHOD == 'linear':
            targets = torch.clamp(targets, min=0., max=1.)

        # Frame resolution sequence mask
        mask = emphases.model.mask_from_lengths(frame_lengths)

    else:

        # Word resolution sequence mask
        mask = emphases.model.mask_from_lengths(word_lengths)

    # Compute masked loss
    if loss_fn == 'bce':
        return torch.nn.functional.binary_cross_entropy_with_logits(
            scores[mask],
            targets[mask])
    elif loss_fn == 'mse':
        return torch.nn.functional.mse_loss(scores[mask], targets[mask])
    raise ValueError(f'Loss {loss_fn} is not recognized')



import os
from pathlib import Path

import torch


###############################################################################
# Metadata
###############################################################################


# Configuration name
CONFIG = 'emphases'


###############################################################################
# Directories
###############################################################################


# Location to save assets to be bundled with pip release
ASSETS_DIR = Path(__file__).parent.parent / 'assets'

# Location of preprocessed features
CACHE_DIR = Path(__file__).parent.parent.parent / 'data' / 'cache'

# Location of datasets on disk
DATA_DIR = Path(__file__).parent.parent.parent / 'data' / 'datasets'

# Location to save evaluation artifacts
EVAL_DIR = Path(__file__).parent.parent.parent / 'eval'

# Location to save training and adaptation artifacts
RUNS_DIR = Path(__file__).parent.parent.parent / 'runs'

# Location of compressed datasets on disk
SOURCE_DIR = Path(__file__).parent.parent.parent / 'data' / 'sources'


###############################################################################
# Audio parameters
###############################################################################


# The maximum representable frequency
FMAX = 550.

# The minumum representable frequency
FMIN = 40.

# The number of samples between frames
HOPSIZE = 160

# Minimum decibel level
MIN_DB = -100.

# Number of linear frequency channels
NUM_FFT =  1024

# Number of mel channels
NUM_MELS = 80

# Voiced/unvoiced threshold for pitch estimation
VOICED_THRESHOLD = .1625

# Reference decibel level
REF_DB = 20.

# The audio samling rate
SAMPLE_RATE = 16000

# The size of the audio analysis window
WINDOW_SIZE = 1024


###############################################################################
# Data parameters
###############################################################################


# List of all datasets
DATASETS = ['libritts']

# Datasets to use for evaluation
EVALUATION_DATASETS = ['libritts']

# Whether to use mel features
MEL_FEATURE = True

# Whether to use loudness features
LOUDNESS_FEATURE = False

# Maximum number of allowed annotations
MAX_ANNOTATIONS = None

# Maximum number of training utterances
MAX_TRAINING_UTTERANCES = None

# Minimum number of allowed annotations
MIN_ANNOTATIONS = None

# Normalize input representations
NORMALIZE = False

# Whether to use the specified one-eighth dataset for scaling law experiments
ONE_EIGHTH_UTTERANCES = False

# Whether to use pitch features
PITCH_FEATURE = False

# Whether to use periodicity features
PERIODICITY_FEATURE = False

# Seed for all random number generators
RANDOM_SEED = 0

# Size of each partition. Must add to 1.
SPLIT_SIZE_TEST = .1
SPLIT_SIZE_TRAIN = .8
SPLIT_SIZE_VALID = .1

# Dataset to use for training
TRAINING_DATASET = 'libritts'

# Dataset to use for validation
VALIDATION_DATASET = 'libritts'


###############################################################################
# Evaluation parameters
###############################################################################


# Number of steps between logging to Tensorboard
LOG_INTERVAL = 100  # steps

# Number of steps to perform for tensorboard logging
LOG_STEPS = 32

# Number of examples to plot to Tensorboard during training
PLOT_EXAMPLES = 2


###############################################################################
# Wavelet baseline parameters
###############################################################################


# Line of maximum amplitude bounds
LOMA_BOUNDARY_START = -2  # octaves
LOMA_BOUNDARY_END = 1  # octaves
LOMA_PROMINENCE_START = -3  # octaves
LOMA_PROMINENCE_END = 0  # octaves

# Weight applied to the duration
PROMINENCE_DURATION_WEIGHT = .5

# Maximum frequency in energy calculation
PROMINENCE_ENERGY_MAX = 5000.

# Minimum frequency in energy calculation
PROMINENCE_ENERGY_MIN = 200.

# Weight applied to the energy
PROMINENCE_ENERGY_WEIGHT = 1.

# Weight applied to the pitch
PROMINENCE_PITCH_WEIGHT = 1.

# Voiced/unvoiced threshold from 0 (all voiced) to 100 (all unvoiced)
VOICED_THRESHOLD = 50


###############################################################################
# Model parameters
###############################################################################


# Activation function to use in convolution model
ACTIVATION_FUNCTION = torch.nn.ReLU

# Model architecture. One of ['convolution', 'transformer'].
ARCHITECTURE = 'convolution'

# Model width
CHANNELS = 80

# Decoder convolution kernel size
DECODER_KERNEL_SIZE = 3

# Dropout probability (or None to not use dropout)
DROPOUT = None

# Location to perform resampling from frame resolution to word resolution.
# One of ['inference', 'input', 'intermediate', 'loss'].
DOWNSAMPLE_LOCATION = 'intermediate'

# Method to use for resampling from frame resolution to word resolution.
# One of ['average', 'center', 'max', 'sum'].
DOWNSAMPLE_METHOD = 'sum'

# Encoder convolution kernel size
ENCODER_KERNEL_SIZE = 3

# Number of network layers
LAYERS = 6

# Method to use for inference. One of
# ['neural', 'pitch-variance', 'duration-variance', 'prominence].
METHOD = 'neural'

# Method to use for resampling from word resolution to frame resolution.
# One of ['linear', 'nearest'].
UPSAMPLE_METHOD = 'linear'


###############################################################################
# Training parameters
###############################################################################


# Number of buckets of data lengths used by the sampler
BUCKETS = 2

# Loss function. One of ['bce', 'mse']
LOSS = 'bce'

# Maximum number of frames in one batch
MAX_TRAINING_FRAMES = 75000

# Number of training steps
NUM_STEPS = 6000

# Number of data loading worker threads
try:
    NUM_WORKERS = int(os.cpu_count() / max(1, len(GPUtil.getGPUs())))
except ValueError:
    NUM_WORKERS = os.cpu_count()



import torch


###############################################################################
# Duration variance baseline method
###############################################################################


def infer(alignment):
    """Compute per-word emphasis scores using duration variance method"""
    # Average duration of phonemes in the sentence
    average_duration = alignment.duration() / len(alignment.phonemes())

    # Average duration of phonemes in each word
    average_duration_per_word = torch.tensor([
        word.duration() / len(word) for word in alignment])

    # Zero-center
    return (average_duration_per_word - average_duration)[None]



import argparse

import emphases


def parse_args():
    """Parse command-line arguments"""
    parser = argparse.ArgumentParser(description='Partition datasets')
    parser.add_argument(
        '--datasets',
        nargs='+',
        default=emphases.DATASETS,
        help='The datasets to partition')
    return parser.parse_known_args()[0]


if __name__ == '__main__':
    emphases.partition.datasets(**vars(parse_args()))



import functools

import torch

import emphases


###############################################################################
# Convolution model
###############################################################################


class Convolution(torch.nn.Sequential):

    def __init__(self, kernel_size=emphases.ENCODER_KERNEL_SIZE):
        # Bind common parameters
        conv_fn = functools.partial(
            torch.nn.Conv1d,
            kernel_size=kernel_size,
            padding='same')

        # Layers
        layers = []
        channels = emphases.CHANNELS
        for _ in range(emphases.LAYERS):
            layers.extend((
                conv_fn(channels, channels),
                emphases.ACTIVATION_FUNCTION()))
            if emphases.DROPOUT is not None:
                layers.append(torch.nn.Dropout(emphases.DROPOUT))

        # Register to Module
        super().__init__(*layers)

    # Ignore sequence length parameter needed for Transformer model
    def forward(self, x, _):
        return super().forward(x)



import multiprocessing as mp
import os

import librosa
import torch
import torchutil

import emphases


###############################################################################
# Mel spectrogram
###############################################################################


def from_audio(audio):
    """Compute spectrogram from audio"""
    # Cache hann window
    if (
        not hasattr(from_audio, 'window') or
        from_audio.dtype != audio.dtype or
        from_audio.device != audio.device
    ):
        from_audio.window = torch.hann_window(
            emphases.WINDOW_SIZE,
            dtype=audio.dtype,
            device=audio.device)
        from_audio.dtype = audio.dtype
        from_audio.device = audio.device

    # Pad audio
    size = (emphases.NUM_FFT - emphases.HOPSIZE) // 2
    audio = torch.nn.functional.pad(
        audio,
        (size, size),
        mode='reflect')

    # Compute stft
    stft = torch.stft(
        audio.squeeze(1),
        emphases.NUM_FFT,
        hop_length=emphases.HOPSIZE,
        window=from_audio.window,
        center=False,
        normalized=False,
        onesided=True,
        return_complex=True)
    stft = torch.view_as_real(stft)[0]

    # Compute magnitude
    spectrogram = torch.sqrt(stft.pow(2).sum(-1) + 1e-6)

    # Convert to mels
    mels = linear_to_mel(spectrogram)

    # Scale to roughly [0, 1]
    if emphases.NORMALIZE:
        return (mels + 10.) / 10.
    return mels


def from_file(audio_file):
    """Load audio and compute mels"""
    audio = emphases.load.audio(audio_file)

    # Compute mels
    return from_audio(audio)


def from_file_to_file(audio_file, output_file):
    """Compute mels from audio file and save to disk"""
    mels = from_file(audio_file)

    # Save to disk
    output_file.parent.mkdir(exist_ok=True, parents=True)
    torch.save(mels, output_file)


def from_files_to_files(audio_files, output_files):
    """Compute mels for many files and save to disk"""
    torchutil.multiprocess_iterator(
        wrapper,
        zip(audio_files, output_files),
        'Preprocessing mels',
        total=len(audio_files),
        num_workers=emphases.NUM_WORKERS)


###############################################################################
# Utilities
###############################################################################


def linear_to_mel(spectrogram):
    # Create mel basis
    if not hasattr(linear_to_mel, 'mel_basis'):
        basis = librosa.filters.mel(
            sr=emphases.SAMPLE_RATE,
            n_fft=emphases.NUM_FFT,
            n_mels=emphases.NUM_MELS)
        basis = torch.from_numpy(basis)
        basis = basis.to(spectrogram.dtype).to(spectrogram.device)
        linear_to_mel.basis = basis

    # Convert to mels
    melspectrogram = torch.matmul(linear_to_mel.basis, spectrogram)

    # Apply dynamic range compression
    return torch.log(torch.clamp(melspectrogram, min=1e-5))

def wrapper(item):
    """Multiprocessing wrapper"""
    from_file_to_file(*item)



import emphases


###############################################################################
# Time conversions
###############################################################################


def frames_to_samples(frames):
    """Convert number of frames to samples"""
    return frames * emphases.HOPSIZE


def frames_to_seconds(frames):
    """Convert number of frames to seconds"""
    return frames * emphases.HOPSIZE_SECONDS


def seconds_to_frames(seconds):
    """Convert seconds to number of frames"""
    return samples_to_frames(seconds_to_samples(seconds))


def seconds_to_samples(seconds):
    """Convert seconds to number of samples"""
    return seconds * emphases.SAMPLE_RATE


def samples_to_frames(samples):
    """Convert samples to number of frames"""
    return samples // emphases.HOPSIZE


def samples_to_seconds(samples):
    """Convert number of samples to seconds"""
    return samples / emphases.SAMPLE_RATE



import numpy as np

import emphases


def rolling_window(a, window):
    shape = a.shape[:-1] + (a.shape[-1] - window + 1, window)
    strides = a.strides + (a.strides[-1],)
    return np.lib.stride_tricks.as_strided(a, shape=shape, strides=strides)


def _cut_boundary_vals(params, num_vals):
    cutted = np.array(params)
    for i in range(num_vals, len(params) - num_vals):
        if params[i] <= 0 and params[i + 1] > 0:
            for j in range(i, i + num_vals):
                cutted[j] = 0.

        if params[i] > 0 and params[i + 1] <= 0:
            for j in range(i - num_vals, i + 1):
                cutted[j] = 0.

    return cutted


def _remove_outliers(log_pitch):
    fixed = np.array(log_pitch)

    # Remove outlier f0 values from voicing boundaries
    boundary_cut = emphases.baselines.prominence.smooth_and_interp.interpolate_zeros(
        _cut_boundary_vals(fixed, 3),
        'linear')
    interp = emphases.baselines.prominence.smooth_and_interp.interpolate_zeros(fixed, 'linear')
    fixed[abs(interp - boundary_cut) > .1] = 0
    interp = emphases.baselines.prominence.smooth_and_interp.interpolate_zeros(fixed, 'linear')

    # iterative outlier removal
    # 1. compare current contour estimate to a smoothed contour and remove deviates larger than threshold
    # 2. smooth current estimate with shorter window, thighten threshold
    # 3. goto 1.

    # In practice, first handles large scale octave jump type errors,
    # finally small scale 'errors' like consonant perturbation effects and
    # other irregularities in voicing boundaries
    #
    # if this appears to remove too many correct values, increase thresholds
    num_iter = 30
    max_win_len = 100
    min_win_len = 10
    max_threshold = 3.  # threshold with broad window
    min_threshold = .5  # threshold with shorted window

    _std = np.std(interp)
    # do not tie fixing to liveliness of the original
    _std = .3

    win_len = np.exp(
        np.linspace(np.log(max_win_len), np.log(min_win_len), num_iter + 1))
    outlier_threshold = np.linspace(
        _std * max_threshold,
        _std * min_threshold,
        num_iter + 1)
    for i in range(0, num_iter):
        smooth_contour = emphases.baselines.prominence.smooth_and_interp.smooth(interp, win_len[i])
        low_limit = smooth_contour - outlier_threshold[i]
        # bit more careful upwards, not to cut emphases
        hi_limit = smooth_contour + outlier_threshold[i] * 1.5

        # octave jump down fix, more harm than good?
        fixed[interp > hi_limit] = 0
        fixed[interp < low_limit] = 0
        interp = emphases.baselines.prominence.smooth_and_interp.interpolate_zeros(fixed, 'linear')

    return fixed


def _interpolate(f0):
    interp = emphases.baselines.prominence.smooth_and_interp.interpolate_zeros(f0)
    _std = np.std(interp)
    _min = np.min(interp)
    low_limit = emphases.baselines.prominence.smooth_and_interp.smooth(interp, 200) - 1.5 * _std
    low_limit[low_limit < _min] = _min
    hi_limit = emphases.baselines.prominence.smooth_and_interp.smooth(interp, 100) + 2. * _std
    voicing = np.array(f0)
    constrained = np.array(f0)
    constrained = np.maximum(f0, low_limit)
    constrained = np.minimum(constrained, hi_limit)
    interp = emphases.baselines.prominence.smooth_and_interp.peak_smooth(
        constrained,
        100,
        20,
        voicing=voicing)
    # smooth voiced parts a bit too
    return emphases.baselines.prominence.smooth_and_interp.peak_smooth(interp, 3, 2)


def process(f0):
    log_pitch = np.array(f0)
    log_scaled = True
    if np.mean(f0[f0 > 0]) > 20:
        log_scaled = False
        log_pitch[f0 > 0] = np.log(f0[f0 > 0])
        log_pitch[f0 <= 0] = 0

    log_pitch = _remove_outliers(log_pitch)
    log_pitch = _interpolate(log_pitch)

    if not log_scaled:
        return np.exp(log_pitch)
    else:
        return log_pitch



import json

import torch
import torchutil

import emphases


###############################################################################
# Evaluate
###############################################################################


@torchutil.notify('evaluate')
def datasets(datasets, checkpoint=None, gpu=None):
    """Perform evaluation"""
    device = torch.device('cpu' if gpu is None else f'cuda:{gpu}')

    # Containers for results
    overall, granular = {}, {}

    # Evaluate each dataset
    for dataset in datasets:

        # Get data loader
        loader = emphases.data.loader(dataset, 'test', gpu)

        # Get mean and variance for Pearson Correlation
        target_stats = emphases.evaluate.metrics.Statistics()
        predicted_stats = emphases.evaluate.metrics.Statistics()
        for batch in loader:

            # Unpack
            _, _, _, word_lengths, targets, alignments, audio, _ = batch

            # Get predicted scores
            scores = emphases.from_alignment_and_audio(
                alignments[0],
                audio[0],
                emphases.SAMPLE_RATE,
                checkpoint=checkpoint,
                gpu=gpu)

            # Update statistics
            target_stats.update(targets, word_lengths)
            predicted_stats.update(scores[None], word_lengths)

        # Get metric class
        metric_fn = emphases.evaluate.Metrics

        # Per-file metrics
        file_metrics = metric_fn(predicted_stats, target_stats)

        # Per-dataset metrics
        dataset_metrics = metric_fn(predicted_stats, target_stats)

        # Iterate over test set
        for batch in torchutil.iterator(
            loader,
            f'Evaluating {emphases.CONFIG} on {dataset}',
            total=len(loader)
        ):

            # Unpack
            (
                _,
                frame_lengths,
                word_bounds,
                word_lengths,
                targets,
                alignments,
                audio,
                stems
             ) = batch

            # Reset file metrics
            file_metrics.reset()

            if emphases.METHOD == 'neural':

                # Get predicted scores
                scores = []

                # Preprocess audio
                for features, word_bounds in emphases.preprocess(
                    alignments[0],
                    audio[0],
                    gpu=gpu
                ):

                    # Infer
                    logits = emphases.infer(
                        features,
                        word_bounds,
                        checkpoint).detach()

                    # Skip postprocessing
                    scores.append(logits)

                # Concatenate results
                scores = torch.cat(scores, 2)

            else:

                # Baseline method inference
                scores = emphases.from_alignment_and_audio(
                    alignments[0],
                    audio[0],
                    emphases.SAMPLE_RATE,
                    gpu=gpu)[None]

            # Update metrics
            args = (scores, targets.to(device), word_lengths.to(device))
            file_metrics.update(*args)
            dataset_metrics.update(*args)

            # Copy results
            granular[f'{dataset}/{stems[0]}'] = file_metrics()
        overall[dataset] = dataset_metrics()

    # Write to json files
    directory = emphases.EVAL_DIR / emphases.CONFIG
    directory.mkdir(exist_ok=True, parents=True)
    with open(directory / 'overall.json', 'w') as file:
        json.dump(overall, file, indent=4)
    with open(directory / 'granular.json', 'w') as file:
        json.dump(granular, file, indent=4)



import torch

import emphases


###############################################################################
# Files and directories
###############################################################################


# Directory to save annotation artifacts
ANNOTATION_DIR = emphases.SOURCE_DIR / 'crowdsource'

# Default configuration file for emphasis annotation
DEFAULT_ANNOTATION_CONFIG = emphases.ASSETS_DIR / 'configs' / 'annotate.yaml'

# Location to save dataset partitions
PARTITION_DIR = emphases.ASSETS_DIR / 'partitions'


###############################################################################
# Audio parameters
###############################################################################


# The hopsize in seconds
HOPSIZE_SECONDS = emphases.HOPSIZE / emphases.SAMPLE_RATE

# The maximum representable frequency in log-hz
LOGFMAX = torch.log2(torch.tensor(emphases.FMAX))

# The minumum representable frequency in log-hz
LOGFMIN = torch.log2(torch.tensor(emphases.FMIN))


###############################################################################
# Model parameters
###############################################################################


# Number of input features to the model
NUM_FEATURES = (
    emphases.MEL_FEATURE * emphases.NUM_MELS +
    int(emphases.PITCH_FEATURE) +
    int(emphases.PERIODICITY_FEATURE) +
    int(emphases.LOUDNESS_FEATURE))



import torch
import torchutil

import emphases


###############################################################################
# Aggregate metric
###############################################################################


class Metrics:

    def __init__(self, predicted_stats, target_stats):
        self.correlation = torchutil.metrics.PearsonCorrelation(
            *predicted_stats(),
            *target_stats())
        self.bce = BinaryCrossEntropy()
        self.mse = MeanSquaredError()

    def __call__(self):
        return {
            'pearson_correlation': self.correlation(),
            'bce': self.bce(),
            'mse': self.mse()}

    def update(
        self,
        logits,
        targets,
        word_lengths):
        # Detach from graph
        logits = logits.detach()

        # Word resolution sequence mask
        mask = emphases.model.mask_from_lengths(word_lengths)
        logits, targets = logits[mask], targets[mask]

        # Update cross entropy
        self.bce.update(logits, targets)

        # Update squared error
        self.mse.update(emphases.postprocess(logits), targets)

        # Update pearson correlation
        self.correlation.update(emphases.postprocess(logits), targets)

    def reset(self):
        self.correlation.reset()
        self.bce.reset()
        self.mse.reset()


###############################################################################
# Individual metrics
###############################################################################


class BinaryCrossEntropy(torchutil.metrics.Average):

    def update(self, scores, targets):
        if emphases.LOSS == 'bce':

            # Get values from logits
            values = torch.nn.functional.binary_cross_entropy_with_logits(
                scores,
                targets,
                reduction='none')

        else:

            # Get values from probabilities
            x, y = torch.clamp(scores, 0., 1.), targets
            values = -(
                y * torch.log(x + 1e-6) + (1 - y) * torch.log(1 - x + 1e-6))

        # Update
        super().update(values, values.numel())


# TODO - fix scaling
class MeanSquaredError(torchutil.metrics.Average):

    def update(
        self,
        scores,
        targets):
        # Compute sum of MSE
        values = torch.nn.functional.mse_loss(
            scores,
            targets,
            reduction='none')

        # Update
        super().update(values, values.numel())


###############################################################################
# Utilities
###############################################################################


class Statistics(torchutil.metrics.MeanStd):

    def update(self, values, lengths):
        # Sequence mask
        mask = emphases.model.mask_from_lengths(lengths)

        # Update
        super().update(values[mask].flatten().tolist())



import argparse
import shutil
from pathlib import Path

import torchutil

import emphases


###############################################################################
# Entry point
###############################################################################


def main(config, dataset, gpu=None):
    # Create output directory
    directory = emphases.RUNS_DIR / config.stem
    directory.mkdir(parents=True, exist_ok=True)

    # Save configuration
    shutil.copyfile(config, directory / config.name)

    # Train
    emphases.train(dataset, directory, gpu)

    # Get best checkpoint
    checkpoint = torchutil.checkpoint.best_path(directory)[0]

    # Evaluate
    emphases.evaluate.datasets(emphases.EVALUATION_DATASETS, checkpoint, gpu)


def parse_args():
    """Parse command-line arguments"""
    parser = argparse.ArgumentParser(description='Train a model')
    parser.add_argument(
        '--config',
        type=Path,
        help='The configuration file')
    parser.add_argument(
        '--dataset',
        default=emphases.TRAINING_DATASET,
        help='The dataset to train on')
    parser.add_argument(
        '--gpu',
        type=int,
        help='The gpu to run training on')
    return parser.parse_args()


if __name__ == '__main__':
    main(**vars(parse_args()))



import fractions

import torch
import numpy as np
from scipy.signal import resample_poly

import emphases


###############################################################################
# Prominence API
###############################################################################


def infer(alignment, audio, sample_rate):
    """Compute per-word prominence from alignment and audio"""
    # Convert to numpy
    audio = audio.numpy()[0]

    # Compute energy
    energy = emphases.baselines.prominence.energy_processing.extract_energy(
        audio,
        sample_rate)
    energy = np.cbrt(energy + 1)

    # Smooth energy
    energy = emphases.baselines.prominence.smooth_and_interp.peak_smooth(
        energy,
        30,
        3)
    energy = emphases.baselines.prominence.smooth_and_interp.smooth(energy, 10)

    # Compute pitch
    pitch = emphases.baselines.prominence.pitch_tracker.inst_freq_pitch(
        audio,
        sample_rate)
    pitch = emphases.baselines.prominence.f0_processing.process(pitch)

    # Extract duration
    duration = \
        emphases.baselines.prominence.duration_processing.get_duration_signal(
        alignment,
        weights=[.5, .5],
        rate=200)

    # Slice features
    min_length = np.min([len(pitch), len(energy), len(duration)])
    pitch = pitch[:min_length]
    energy = energy[:min_length]
    duration = duration[:min_length]

    # Combine features
    combined = (
        emphases.PROMINENCE_PITCH_WEIGHT * normalize(pitch) +
        emphases.PROMINENCE_ENERGY_WEIGHT * normalize(energy) +
        emphases.PROMINENCE_DURATION_WEIGHT * normalize(duration))
    combined = normalize(
        emphases.baselines.prominence.smooth_and_interp.remove_bias(
            combined,
            800))

    # Distance between adjacent scales (.25 means 4 scales per octave)
    scale_distance = .25  # octaves

    # Continuous wavelet transform analysis
    cwt, scales, freqs = emphases.baselines.prominence.cwt_utils.cwt_analysis(
        combined,
        mother_name='mexican_hat',
        period=3,
        num_scales=34,
        scale_distance=scale_distance,
        apply_coi=False)
    cwt = np.real(cwt)
    scales *= 200

    # Get scale that minimizes distance with average word length
    average_duration = (alignment.end() / len(alignment))*200
    scales = 1. / freqs * 200 * .5
    scale = np.argmin(np.abs(scales - average_duration))

    # Define the scale information
    pos_loma_start = scale + \
        int(emphases.LOMA_PROMINENCE_START / scale_distance)
    pos_loma_end = scale + \
        int(emphases.LOMA_PROMINENCE_END / scale_distance)
    neg_loma_start = scale + \
        int(emphases.LOMA_BOUNDARY_START / scale_distance)
    neg_loma_end = scale + \
        int(emphases.LOMA_BOUNDARY_END / scale_distance)

    # Retrieve line of maximum amplitude
    pos_loma = emphases.baselines.prominence.loma.get_loma(
        cwt,
        scales,
        pos_loma_start,
        pos_loma_end)
    neg_loma = emphases.baselines.prominence.loma.get_loma(
        -cwt,
        scales,
        neg_loma_start,
        neg_loma_end)

    # Decode prominence
    max_loma = np.array(emphases.baselines.prominence.loma.get_prominences(
        pos_loma,
        alignment,
        rate=200))

    # Prominence dimensions - [time, value]
    prominences = torch.tensor(max_loma)

    # Decode boundaries
    # Boundries dimensions - [time, value]
    boundaries = torch.tensor(emphases.baselines.prominence.loma.get_boundaries(
        max_loma,
        neg_loma,
        alignment))

    return prominences[:, 1][None]


###############################################################################
# Utilities
###############################################################################


def normalize(features):
    """Normalize features"""
    return (features - np.nanmean(features)) / (np.nanstd(features) + 1e-7)


def resample(signal, original_sample_rate, target_sample_rate):
    """Resample signal"""
    ratio = fractions.Fraction(target_sample_rate, original_sample_rate)
    return resample_poly(signal, ratio.numerator, ratio.denominator)



import csv
import json
import shutil
import ssl
import tarfile
import urllib
import yaml

import pyfoal
import pypar
import torch
import torchutil
import torchaudio

import emphases


###############################################################################
# Constants
###############################################################################


# List of tokens to filter from Buckeye annotations
BUCKEYE_FILTER_LIST = [
    '{B_TRANS}',
    '{E_TRANS}',
    '<CUTOFF-i=?>',
    '<CUTOFF-ta=taking?>',
    '<CUTOFF-th=that>',
    '<IVER>',
    '<LAUGH>',
    '<SIL>',
    '<VOCNOISE>',
    '<and>',
    '<i>',
    '<out>',
    '<that>',
    '<think>',
    '<so>',
    '<some>',
    '<um>',
    '<xx>',
]

# Speakers selected by sorting the train-clean-100 speakers by longest total
# recording duration and manually selecting speakers with more natural,
# conversational (as opposed to read) prosody
LIBRITTS_SPEAKERS = [
    # Top 5 Female (primarily by length)
    40,
    669,
    4362,
    5022,
    8123,

    # Additional female speakers to get to 1/8th of train-clean-100
    5022,
    696,
    6272,
    5163,

    # Top 5 Male (primarily by length)
    196,
    460,
    1355,
    3664,
    7067,

    # Additional male speakers to get to 1/8th of train-clean-100
    405,
    6437,
    446,
    4397
]


###############################################################################
# Download datasets
###############################################################################


@torchutil.notify('download')
def datasets(datasets, gpu=None):
    """Download datasets"""
    for dataset in datasets:
        if dataset == 'automatic':
            automatic(gpu=gpu)
        elif dataset == 'buckeye':
            buckeye()
        elif dataset == 'crowdsource':
            crowdsource()
        elif dataset == 'libritts':
            libritts()
        else:
            raise ValueError(f'Dataset {dataset} is not defined')


###############################################################################
# Individual dataset downloaders
###############################################################################


def automatic(gpu=None):
    """Create dataset from trained model"""
    # Setup directories
    cache_directory = emphases.CACHE_DIR / 'automatic'
    cache_directory.mkdir(exist_ok=True, parents=True)

    # Create subdirectories
    features = ['alignment', 'audio', 'scores']
    for feature in features:
        (cache_directory / feature).mkdir(exist_ok=True, parents=True)

    # Get files
    audio_files = list(
        (emphases.CACHE_DIR / 'libritts' / 'audio').rglob('*.wav'))
    stems = [file.stem for file in audio_files]

    # Copy from LibriTTS cache to annotation cache
    for stem in stems:

        # Copy audio
        audio_file = (
            emphases.CACHE_DIR / 'automatic' / 'audio' / f'{stem}.wav')
        shutil.copyfile(
            emphases.CACHE_DIR / 'libritts' / 'audio' / f'{stem}.wav',
            audio_file)

        # Copy alignment
        shutil.copyfile(
            emphases.CACHE_DIR / 'libritts' / 'alignment' / f'{stem}.TextGrid',
            emphases.CACHE_DIR / 'automatic' / 'alignment' / f'{stem}.TextGrid')

        # Load alignment
        alignment = pypar.Alignment(
            emphases.CACHE_DIR / 'automatic' / 'alignment' / f'{stem}.TextGrid')

        # Load audio
        audio, _ = torchaudio.load(audio_file)

        # Infer scores
        scores = emphases.from_alignment_and_audio(
            alignment,
            audio,
            emphases.SAMPLE_RATE,
            gpu=gpu).detach().cpu()

        # Save scores
        torch.save(scores, cache_directory / 'scores' / f'{stem}.pt')


def crowdsource():
    """Prepare crowdsourced dataset"""
    # Get annotation config
    with open(emphases.DEFAULT_ANNOTATION_CONFIG, "r") as stream:
        annotation_config = yaml.safe_load(stream)

    # Setup directories
    data_directory = emphases.DATA_DIR / 'crowdsource'
    cache_directory = emphases.CACHE_DIR / 'crowdsource'
    cache_directory.mkdir(exist_ok=True, parents=True)

    # Create subdirectories
    features = ['alignment', 'audio', 'scores']
    for feature in features:
        (cache_directory / feature).mkdir(exist_ok=True, parents=True)

    # Load annotations data
    annotation_data = {}
    for directory in data_directory.glob('*'):

        source_directory =  directory / annotation_config['name']
        table_directory = source_directory / 'tables'

        # Participant data
        participants = {}
        with open(table_directory / 'participants.csv') as file:
            for row in csv.DictReader(file):
                try:

                    # Crowdsourced annotation
                    participants[row['ID']] = {
                        'language': row['Language'],
                        'country': row['Country'],
                        'annotations': []}

                except KeyError as error:

                    # Manual annotation
                    participants[row['ID']] = {
                        'language': 'English',
                        'country': 'United States',
                        'annotations': []}

        # Response data
        with open(table_directory / 'responses.csv') as file:
            for row in csv.DictReader(file):
                participant = row['Participant']

                # Add participant
                if participant not in annotation_data:
                    annotation_data[participant] = participants[participant]

                # Get word start and end times
                alignment = pypar.Alignment(
                    emphases.CACHE_DIR /
                    'libritts' /
                    'alignment' /
                    f'{row["Stem"]}.TextGrid')
                words = [
                    (str(word).lower(), word.start(), word.end())
                    for word in alignment
                    if str(word) != pypar.SILENCE]

                # Format annotation
                entry = {
                    'stem': row['Stem'],
                    'score': [float(c) for c in row['Response']],
                    'words': words}
                assert len(entry['words']) == len(entry['score'])

                # Add annotation
                annotation_data[participant]['annotations'].append(entry)

    # Get worker ID correspondence
    correspondence = {}
    for directory in data_directory.glob('*'):
        file = (
            directory /
            annotation_config['name'] /
            'crowdsource' /
            'crowdsource.json')
        with open(file) as file:
            contents = json.load(file)
            for content in contents:
                correspondence |= {content['ParticipantID']: content['WorkerId']}

    # Crowdsourced annotation
    if correspondence:

        # Filter out where incomplete or > 1/3 examples have > 2/3 words selected
        def valid(items):
            if not hasattr(valid, 'count'):
                valid.count = 0
            sums = [sum(item['score']) for item in items]
            counts = [len(item['score']) for item in items]
            invalids = [s > .67 * c for s, c in zip(sums, counts)]
            is_valid = sum(invalids) < .33 * len(invalids)
            valid.count += 1 - int(is_valid)
            return is_valid

        # Join participants with same worker ID
        joined = {}
        for participant, contents in annotation_data.items():

            # Filter out bad batches
            if (
                len(contents['annotations']) < 20 or
                len(contents['annotations']) % 10 > 0 or
                not valid(contents['annotations'])
            ):
                continue

            worker = correspondence[participant]
            if worker in joined:
                joined[worker]['annotations'].extend(contents['annotations'])
            else:
                joined[worker] = contents

    # Manual annotation
    else:
        joined = annotation_data

    # Anonymize
    anonymized = {}
    for i, contents in enumerate(joined.values()):
        anonymized[f'{i:06d}'] = contents

    # Save annotations in release format
    with open(cache_directory / 'annotations.json', 'w') as file:
        json.dump(anonymized, file, sort_keys=True, indent=True)

    # Merge binary annotations to floats
    annotations = merge_annotations(anonymized)

    # Save dictionary containing annotation counts
    with open(cache_directory / 'counts.json', 'w') as file:
        json.dump(annotations['stems'], file, sort_keys=True, indent=True)

    # Get annotated stems
    stems = [
        file.replace('libritts-', '')
        for file in annotations['stems'].keys()]

    # Copy from LibriTTS cache to annotation cache
    for i, stem in enumerate(stems):

        # Get normalized scores
        count = annotations['stems'][stem]
        labels = [score / count for score in annotations['scores'][stem]]

        # Copy audio
        shutil.copyfile(
            emphases.CACHE_DIR / 'libritts' / 'audio' / f'{stem}.wav',
            emphases.CACHE_DIR / 'crowdsource' / 'audio' / f'{stem}.wav')

        # Copy alignment
        shutil.copyfile(
            emphases.CACHE_DIR / 'libritts' / 'alignment' / f'{stem}.TextGrid',
            emphases.CACHE_DIR / 'crowdsource' / 'alignment' / f'{stem}.TextGrid')

        # Load alignment
        alignment = pypar.Alignment(
            emphases.CACHE_DIR / 'crowdsource' / 'alignment' / f'{stem}.TextGrid')

        # Match alignment and scores (silences get a score of zero)
        j = 0
        scores = torch.zeros(len(alignment))
        for i, word in enumerate(alignment):

            # Keep silences as zero
            if str(word) == pypar.SILENCE:
                continue

            # Update scores
            scores[i] = float(labels[j])

            j += 1

        # Save scores
        torch.save(scores, cache_directory / 'scores' / f'{stem}.pt')


def buckeye():
    """Download buckeye dataset"""
    # Extract tar file to data directory
    file = emphases.SOURCE_DIR / 'buckeye' / 'buckeye.tar.gz'
    with tarfile.open(file, 'r:gz') as tfile:
        tfile.extractall(emphases.DATA_DIR)

    # Setup cache directory
    cache_directory = emphases.CACHE_DIR / 'buckeye'
    cache_directory.mkdir(exist_ok=True, parents=True)

    # Create subdirectories
    features = ['alignment', 'audio', 'scores']
    for feature in features:
        (cache_directory / feature).mkdir(exist_ok=True, parents=True)

    # Copy alignments and filter out unused tokens
    data_directory = emphases.DATA_DIR / 'buckeye'
    alignment_files = (data_directory / 'alignment').glob('*.TextGrid')
    for file in alignment_files:

        # Load alignment
        alignment = pypar.Alignment(file)

        # Filter
        for word in alignment:
            if str(word) in BUCKEYE_FILTER_LIST:
                word.word = pypar.SILENCE
                word.phonemes = [
                    pypar.Phoneme(pypar.SILENCE, word.start(), word.end())]

        # Deduplicate silence tokens
        i = 0
        words = alignment.words()
        prev_silence = False
        while i < len(words):
            word = words[i]
            if str(word) == pypar.SILENCE:
                if prev_silence:
                    words[i - 1][-1]._end = word.end()
                    del words[i]
                else:
                    prev_silence = True
                    i += 1
            else:
                prev_silence = False
                i += 1

        # Save alignment
        pypar.Alignment(words).save(
            cache_directory / 'alignment' / f'{file.stem}.TextGrid')

    # Get audio files
    audio_files = sorted((data_directory / 'audio').glob('*.wav'))

    # Resample audio
    for audio_file in audio_files:

        # Load and resample
        audio = emphases.load.audio(audio_file)

        # If audio is too quiet, increase the volume
        maximum = torch.abs(audio).max()
        if maximum < .35:
            audio *= .35 / maximum

        # Save to disk
        torchaudio.save(
            cache_directory / 'audio' / audio_file.name,
            audio,
            emphases.SAMPLE_RATE)

    # Read buckeye annotations
    data_directory = emphases.DATA_DIR / 'buckeye'
    with open(data_directory  / 'annotations.csv') as file:
        reader = csv.DictReader(file)
        annotations = [row for row in reader]

    # Extract per-word emphasis scores
    alignment_files = (cache_directory / 'alignment').glob('*.TextGrid')
    for file in alignment_files:

        # Load alignment
        alignment = pypar.Alignment(file)

        # Get words from annotation
        words = [word for word in annotations if word['filename'] == file.stem]
        words = sorted(words, key=lambda x: float(x['wordmin']))

        # Get per-word emphasis scores
        j = 0
        scores = torch.zeros(len(alignment))
        for i, word in enumerate(alignment):

            # Keep silences as zero
            if str(word) == pypar.SILENCE:
                continue

            # Make sure alignments are aligned
            assert str(word).lower() == words[j]['word'].lower()
            assert (word.start() - float(words[j]['wordmin'])) < 1e-4
            assert (word.end() - float(words[j]['wordmax'])) < 1e-4

            # Update scores
            # pa.32 is the average of 32 human judgments of the perception of
            # prominence based on acoustic features
            scores[i] = float(words[j]['pa.32'])

            j += 1

        # Save scores
        torch.save(scores, cache_directory / 'scores' / f'{file.stem}.pt')


def libritts():
    """Download libritts dataset"""
    # Setup source directory
    source_directory = emphases.SOURCE_DIR / 'libritts'
    source_directory.mkdir(exist_ok=True, parents=True)

    # Download
    url = 'https://us.openslr.org/resources/60/train-clean-100.tar.gz'
    file = source_directory / 'libritts-train-clean-100.tar.gz'
    torchutil.download.file(url, file)

    # Unzip
    with tarfile.open(file, 'r:gz') as tfile:
        tfile.extractall(emphases.DATA_DIR)

    # Rename folder
    directory = emphases.DATA_DIR / 'libritts'
    shutil.rmtree(directory, ignore_errors=True)
    shutil.move(emphases.DATA_DIR / 'LibriTTS', directory)

    # Download annotations from zenodo
    url = 'https://zenodo.org/records/10402793/files/libritts-emphasis-annotations.json?download=1'
    file = source_directory / 'annotations.json'
    torchutil.download.file(url, file)

    # Load annotations
    with open(source_directory / 'annotations.json') as file:
        annotations = json.load(file)

    # Merge annotations to floats
    annotations = merge_annotations(annotations)

    # Get list of audio files
    audio_files = list(directory.rglob('*.wav'))
    audio_files = [
        file for file in audio_files if file.stem in annotations['stems']]

    # Setup cache directory
    cache_directory = emphases.CACHE_DIR / 'libritts'
    cache_directory.mkdir(exist_ok=True, parents=True)

    # Create subdirectories
    features = ['alignment', 'audio', 'scores']
    for feature in features:
        (cache_directory / feature).mkdir(exist_ok=True, parents=True)

    # Iterate over files
    for audio_file in torchutil.iterator(
        audio_files,
        'Formatting libritts',
        total=len(audio_files)
    ):

        # Load and resample audio
        audio = emphases.load.audio(audio_file)

        # If audio is too quiet, increase the volume
        maximum = torch.abs(audio).max()
        if maximum < .35:
            audio *= .35 / maximum

        # Save audio
        stem = audio_file.stem
        torchaudio.save(
            cache_directory / 'audio' / f'{stem}.wav',
            audio,
            emphases.SAMPLE_RATE)

    # Align text and audio
    text_files = [
        file.with_suffix('.normalized.txt') for file in audio_files]
    alignment_files = [
        cache_directory / 'alignment' / f'{file.stem}.TextGrid'
        for file in audio_files]
    pyfoal.from_files_to_files(
        text_files,
        audio_files,
        alignment_files,
        'p2fa')

    for i, stem in enumerate([file.stem for file in audio_files]):

        # Load alignment
        alignment = pypar.Alignment(
            cache_directory / 'alignment' / f'{stem}.TextGrid')

        # Get ground truth
        count = annotations['stems'][stem]
        labels = [score / count for score in annotations['scores'][stem]]

        # Match alignment and scores (silences get a score of zero)
        j = 0
        scores = torch.zeros(len(alignment))
        for i, word in enumerate(alignment):

            # Keep silences as zero
            if str(word) == pypar.SILENCE:
                continue

            # Update scores
            scores[i] = float(labels[j])

            j += 1

        # Save scores
        torch.save(scores, cache_directory / 'scores' / f'{stem}.pt')


###############################################################################
# Utilities
###############################################################################


def download_file(url, file):
    """Download file from url"""
    with urllib.request.urlopen(url, context=ssl.SSLContext()) as response, \
            open(file, 'wb') as output:
        shutil.copyfileobj(response, output)


def merge_annotations(annotations):
    """Merge crowdsourced annotations"""
    merged = {'samples': 0, 'scores': {}, 'stems': {}}
    for _, responses in annotations.items():

        # Iterate over stems
        for response in responses['annotations']:
            stem = response['stem']
            score = [float(c) for c in list(response['score'])]

            # Merge stem annotations
            if stem in merged['stems']:

                # Maybe cap the number of allowed annotations
                if (
                    emphases.MAX_ANNOTATIONS is not None and
                    merged['stems'][stem] == emphases.MAX_ANNOTATIONS
                ):
                    continue

                # Update sums and counts
                for i in range(len(score)):
                    merged['scores'][stem][i] += score[i]
                merged['stems'][stem] += 1

            # Add new stem
            else:
                merged['scores'][stem] = score
                merged['stems'][stem] = 1

            # Update total number of samples
            merged['samples'] += 1

    # Maybe cap the minimum required annotations
    if emphases.MIN_ANNOTATIONS is not None:
        merged['stems'] = {
            stem: count for stem, count in merged['stems'].items()
            if count == emphases.MIN_ANNOTATIONS}
        merged['scores'] = {
            stem: scores for stem, scores in merged['scores'].items()
            if stem in merged['stems']}

    return merged



import torch

import emphases


###############################################################################
# Batch collation
###############################################################################


def collate(batch):
    """Batch collation"""
    # Unpack
    features, scores, word_bounds, alignments, audios, stems = zip(*batch)

    # Get word lengths
    word_lengths = torch.tensor(
        [bounds.shape[-1] for bounds in word_bounds],
        dtype=torch.long)
    max_word_length = word_lengths.max().item()

    # Get frame lengths
    frame_lengths = torch.tensor(
        [feat.shape[-1] for feat in features],
        dtype=torch.long)
    max_frame_length = frame_lengths.max().item()

    # Network output lengths
    output_lengths = word_lengths
    max_output_length = max_word_length

    # Allocate padded tensors
    padded_features = torch.zeros(
        (len(features), emphases.NUM_FEATURES, max_frame_length))
    padded_scores = torch.zeros((len(scores), 1, max_output_length))
    padded_bounds = torch.zeros(
        (len(word_bounds), 2, max_word_length),
        dtype=torch.long)
    padded_audio = torch.zeros(
        (len(audios), 1, max_frame_length * emphases.HOPSIZE))

    # Place batch in padded tensors
    for (
        i,
        (bounds, audio, feat, score, frame_length, word_length, output_length)
    ) in enumerate(
        zip(
            word_bounds,
            audios,
            features,
            scores,
            frame_lengths,
            word_lengths,
            output_lengths)
    ):

        # Pad features
        padded_features[i, :, :frame_length] = feat

        # Pad scores
        padded_scores[i, :, :output_length] = score[:, :output_length]

        # Pad word bounds
        padded_bounds[i, :, :word_length] = bounds[:, :word_length]

        # Pad audio
        end_sample = frame_length * emphases.HOPSIZE
        padded_audio[i, :, :end_sample] = audio[:, :end_sample]

    return (
        padded_features,
        frame_lengths,
        padded_bounds,
        word_lengths,
        padded_scores,
        alignments,
        padded_audio,
        stems)



import torch

import emphases


###############################################################################
# Dataloader
###############################################################################


def loader(dataset, partition=None, gpu=None):
    """Retrieve a data loader"""
    # Get dataset
    dataset = emphases.data.Dataset(dataset, partition)

    # Get sampler
    sampler = emphases.data.sampler(dataset, partition)

    # Create loader
    return torch.utils.data.DataLoader(
        dataset,
        num_workers=emphases.NUM_WORKERS,
        pin_memory=gpu is not None,
        collate_fn=emphases.data.collate,
        batch_sampler=sampler)



from numpy import array, sqrt, pad, mean, pi

import pycwt as cwt


###########################################################################################
# Private routines
###########################################################################################


def _padded_cwt(params, dt, dj, s0, J, mother, padding_len):
    """Private function to compute a wavelet transform on padded data

    Parameters
    ----------
    params: arraylike
        The prosodic parameters.
    dt: ?
        ?
    dj: ?
        ?
    s0: ?
        ?
    J: ?
        ?
    mother: ?
        The mother wavelet.
    padding_len: int
        The padding length

    Returns
    -------
    wavelet_matrix: ndarray
    	The wavelet data resulting from the analysis
    scales: arraylike
    	The scale indices corresponding to the wavelet data
    freqs: ?
    	?
    coi: array
    	The cone of influence values
    fft: ?
    	?
    fftfreqs: ?
    	?
    """
    padded = pad(params, padding_len, mode='edge')
    wavelet_matrix, scales, freqs, coi, fft, fftfreqs = cwt.cwt(
        padded,
        dt,
        dj,
        s0,
        J,
        mother)
    wavelet_matrix = \
        wavelet_matrix[:, padding_len:len(wavelet_matrix[0]) - padding_len]
    return wavelet_matrix, scales, freqs, coi, fft, fftfreqs


def _zero_outside_coi(wavelet_matrix, freqs, rate=200):
    """Private function to set each elements outside of the Cone Of Influence (coi) to 0.

    Parameters
    ----------
    wavelet_matrix: type
        description
    freqs: type
        description
    """
    for i in range(0, wavelet_matrix.shape[0]):
        coi = int(1. / freqs[i] * rate)
        wavelet_matrix[i, :coi] = 0.
        wavelet_matrix[i, -coi:] = 0.
    return wavelet_matrix


def _scale_for_reconstruction(
    wavelet_matrix,
    scales,
    dj,
    dt,
    mother='mexican_hat',
    period=3):
    """ ?

    Parameters
    ----------
    wavelet_matrix: ndarray
    	The wavelet data resulting from the analysis
    scales: arraylike
    	The scale indices corresponding to the wavelet data
    dj: ?
        ?
    dt: ?
        ?
    mother: ?
        ?
    period: ?
        ?
    """
    scaled = array(wavelet_matrix)

    # mexican Hat
    c = dj / (3.541 * .867)

    if mother == 'morlet':
        cc = 1.83
        #periods 5 and 6 are correct, 3,4 approximate
        if period == 3:
            cc = 1.74
        if period == 4:
            cc = 1.1
        elif period == 5:
            cc = .9484
        elif period == 6:
            cc = .7784
        c = dj / (cc * pi ** (-.25))

    for i in range(0, len(scales)):
        scaled[i] *= c * sqrt(dt) / sqrt(scales[i])
        # substracting the mean should not be necessary?
        scaled[i] -= mean(scaled[i])

    return scaled


def cwt_analysis(
    params,
    mother_name='mexican_hat',
    num_scales=12,
    first_scale=None,
    scale_distance=1.,
    apply_coi=True,
    period=5,
    frame_rate=200):
    """Achieve the continous wavelet analysis of given parameters

    Parameters
    ----------
    params: arraylike
        The parameters to analyze.
    mother_name: string, optional
        The name of the mother wavelet [default: mexican_hat].
    num_scales: int, optional
        The number of scales [default: 12].
    first_scale: int, optional
        The width of the shortest scale
    scale_distance: float, optional
        The distance between scales [default: 1.0].
    apply_coi: boolean, optional
        Apply the Cone Of Influence (coi)
    period: int, optional
        The period of the mother wavelet [default: 5].
    frame_rate: int, optional
        The signal frame rate [default: 200].

    Returns
    -------
    wavelet_matrix: ndarray
    	The wavelet data resulting from the analysis
    scales: arraylike
    	The scale indices corresponding to the wavelet data
    """
    # setup wavelet transform
    dt = 1. / float(frame_rate)  # frame length

    if not first_scale:
        first_scale = dt # first scale, here frame length

    dj = scale_distance  # distance between scales in octaves
    J = num_scales #  number of scales

    mother = cwt.MexicanHat()

    if str.lower(mother_name) == 'morlet':
        mother = cwt.Morlet(period)

    wavelet_matrix, scales, freqs, *_ = _padded_cwt(
        params,
        dt,
        dj,
        first_scale,
        J,
        mother,
        400)
    wavelet_matrix = _scale_for_reconstruction(
        wavelet_matrix,
        scales,
        dj,
        dt,
        mother=mother_name,
        period=period)

    if apply_coi:
        wavelet_matrix = _zero_outside_coi(wavelet_matrix, freqs, frame_rate)

    return wavelet_matrix, scales, freqs



import json
import random
import shutil

import pypar
import reseval

import emphases


###############################################################################
# Annotate emphases
###############################################################################


def datasets(
    annotation_config=emphases.DEFAULT_ANNOTATION_CONFIG,
    dataset='libritts',
    directory=emphases.ANNOTATION_DIR,
    remote=False,
    production=False,
    interval=120):
    """Perform emphasis annotation on datasets"""
    # Create input and output directories
    directory.mkdir(exist_ok=True, parents=True)
    index = f'{len(list(directory.glob("*"))):02}'
    input_directory = directory / index
    input_directory.mkdir(exist_ok=True, parents=True)
    output_directory = emphases.DATA_DIR / 'crowdsource' / index
    output_directory.mkdir(exist_ok=True, parents=True)

    # Get audio files
    cache_directory = emphases.CACHE_DIR / dataset
    audio_files = sorted(list(cache_directory.rglob('*.wav')))

    # Deterministic shuffle
    random.seed(emphases.RANDOM_SEED)
    random.shuffle(audio_files)

    # Iterate over audio files
    for audio_file in audio_files:

        # Save audio
        shutil.copyfile(audio_file, input_directory / audio_file.name)

        # Load alignment
        alignment = pypar.Alignment(
            cache_directory /
            'alignment' /
            f'{audio_file.stem}.TextGrid')

        # Save text
        text_file = input_directory / f'{audio_file.stem}-words.txt'
        with open(text_file, 'w') as file:
            file.write(
                ' '.join([
                    str(word) for word in alignment
                    if str(word) != pypar.SILENCE]))

    # Run annotation
    reseval.run(
        annotation_config,
        input_directory,
        output_directory,
        not remote,
        production,
        interval)



import matplotlib
import matplotlib.pyplot as plt
import torch

import emphases


###############################################################################
# Plot scaling laws
###############################################################################


def scaling_laws(
    evaluations,
    xlabel,
    output_file,
    yticks,
    scores=None,
    steps=None,
    sizes=None,
    text_offsets=None):
    """Plot scaling laws"""
    # Load evaluation results
    if scores is None or steps is None:
        scores, steps = [], []
        for evaluation in evaluations:
            path, score = emphases.checkpoint.best_path(
                emphases.RUNS_DIR / evaluation)
            checkpoint = torch.load(path, map_location='cpu')
            scores.append(score)
            steps.append(checkpoint['step'])

    # Get x values
    x = [int(eval.split('-')[-1]) for eval in evaluations]

    # Create plot
    figure, axis = plt.subplots(figsize=(8, 2))

    # Remove frame
    axis.spines['top'].set_visible(False)
    axis.spines['right'].set_visible(False)
    axis.spines['bottom'].set_visible(False)
    axis.spines['left'].set_visible(False)

    # Format x axis
    x_range = max(x) - min(x)
    axis.set_xlim([0, max(x) + 0.1 * x_range])
    axis.get_xaxis().set_ticks(x)
    axis.set_xlabel(xlabel)
    axis.xaxis.set_ticks(x)
    axis.xaxis.set_ticklabels(x)

    # Format y axis
    axis.get_yaxis().set_ticks(yticks)
    axis.set_ylim([min(yticks) - .002, max(yticks) + .002])
    axis.tick_params(axis=u'both', which=u'both',length=0)
    axis.set_ylabel('Pearson correlation')

    # Grid lines
    for tick in yticks:
        axis.axhline(tick, color='gray', linestyle='--', linewidth=.8)

    # Plot
    colors = ['blue', 'orange', 'purple', 'red']
    for i in range(len(x)):
        axis.scatter(x[i], scores[i], color=colors[i])

    # Default text offset
    if text_offsets is None:
        text_offsets = [0.011] * len(evaluations)

    # Annotate
    for i in range(len(evaluations)):
        text = f'steps={steps[i]}'
        if sizes is not None:
            text += f'\nutterances={sizes[i]}'
        axis.text(
            x[i],
            scores[i] - text_offsets[i],
            text,
            horizontalalignment='center')

    # Save
    figure.savefig(output_file, bbox_inches='tight', pad_inches=0, dpi=300)



import matplotlib.pyplot as plt

import torch


###############################################################################
# Plot prominence
###############################################################################


def scores(alignment, scores, targets=None):
    """Plot the aligned prominence scores"""
    figure, axis = plt.subplots(figsize=(30, 5))
    axis.set_axis_off()
    axis.set_ylim([0., 1.])

    # Get words, start times, and durations
    centers = [word.start() + word.duration() / 2. for word in alignment]
    duration = [word.duration() for word in alignment]

    # Plot scores
    axis.bar(
        centers,
        scores,
        duration,
        edgecolor='black')

    # Plot words and dividers
    for word in alignment:
        axis.text(
            word.start() + word.duration() / 2,
            .015,
            str(word),
            fontsize=10,
            rotation=90,
            horizontalalignment='center')
        axis.axvline(
            word.start(),
            color='gray',
            linewidth=.5,
            ymin=0.,
            ymax=1.,
            clip_on=False,
            linestyle='--')
    axis.axvline(
        alignment.duration(),
        color='gray',
        linewidth=.5,
        ymin=0.,
        ymax=1.,
        clip_on=False,
        linestyle='--')

    if targets is not None:

        # Plot targets
        axis.bar(centers, targets, duration)

        # Plot overlap
        overlap = torch.minimum(scores, targets)
        axis.bar(centers, overlap, duration, color='gray')

    return figure



from .core import *
from . import cwt_utils
from . import duration_processing
from . import energy_processing
from . import f0_processing
from . import filter
from . import loma
from . import pitch_tracker
from . import smooth_and_interp



from .convolution import Convolution
from .transformer import Transformer

import emphases


def Layers(**kwargs):
    if emphases.ARCHITECTURE == 'convolution':
        return Convolution(**kwargs)
    elif emphases.ARCHITECTURE == 'transformer':
        return Transformer()
    else:
        raise ValueError(
            f'Network layer {emphases.ARCHITECTURE} is not defined')



import penn
import torch
import torchutil

import emphases


###############################################################################
# Preprocess
###############################################################################


@torchutil.notify('preprocess')
def datasets(datasets, gpu=None):
    """Preprocess datasets"""
    for dataset in datasets:
        cache_directory = emphases.CACHE_DIR / dataset

        # Get audio files, from cache
        audio_files = sorted(cache_directory.rglob('*.wav'))

        # Preprocess mels
        mel_files = [
            cache_directory / 'mels' / f'{file.stem}.pt'
            for file in audio_files]
        emphases.data.preprocess.mels.from_files_to_files(
            audio_files,
            mel_files)

        # Preprocess loudness
        loudness_files = [
            cache_directory / 'loudness' / f'{file.stem}.pt'
            for file in audio_files]
        emphases.data.preprocess.loudness.from_files_to_files(
            audio_files,
            loudness_files)

        # Preprocess pitch, periodicity
        (cache_directory / 'pitch').mkdir(exist_ok=True, parents=True)
        pitch_files = [
            cache_directory / 'pitch' / f'{file.stem}'
            for file in audio_files]
        penn.from_files_to_files(
            audio_files,
            pitch_files,
            hopsize=emphases.convert.samples_to_seconds(emphases.HOPSIZE),
            fmin=emphases.FMIN,
            fmax=emphases.FMAX,
            batch_size=2048,
            center='half-hop',
            interp_unvoiced_at=emphases.VOICED_THRESHOLD,
            num_workers=emphases.NUM_WORKERS,
            gpu=gpu)

        # Pitch and periodicity use floating-point hopsize, while mels and
        # loudness use an integer hopsize in samples. This results in
        # single-frame differences when the audio length is within one sample
        # of a new frame due to floating-point error. We simply remove the last
        # frame in this rare case.
        for loudness_file, pitch_file in zip(loudness_files, pitch_files):
            loudness = torch.load(loudness_file)
            pitch = torch.load(f'{pitch_file}-pitch.pt')
            periodicity = torch.load(f'{pitch_file}-periodicity.pt')
            if pitch.shape[1] == loudness.shape[1] + 1:
                pitch = pitch[:, :-1]
                periodicity = periodicity[:, :-1]
            torch.save(pitch, f'{pitch_file}-pitch.pt')
            torch.save(periodicity, f'{pitch_file}-periodicity.pt')


def from_audio(audio, gpu=None):
    """Preprocess one audio file"""
    # Move to device (no-op if devices are the same)
    audio = audio.to('cpu' if gpu is None else f'cuda:{gpu}')

    features = []

    # Preprocess mels
    if emphases.MEL_FEATURE:
        features.append(emphases.data.preprocess.mels.from_audio(audio))

    # Preprocess pitch and periodicity
    if emphases.PITCH_FEATURE or emphases.PERIODICITY_FEATURE:
        pitch, periodicity = penn.from_audio(
            audio,
            emphases.SAMPLE_RATE,
            hopsize=emphases.convert.samples_to_seconds(emphases.HOPSIZE),
            fmin=emphases.FMIN,
            fmax=emphases.FMAX,
            pad=True,
            interp_unvoiced_at=emphases.VOICED_THRESHOLD,
            gpu=gpu)

        if emphases.PITCH_FEATURE:
            if emphases.NORMALIZE:
                features.append(
                    (torch.log2(pitch) - emphases.LOGFMIN) /
                    (emphases.LOGFMAX - emphases.LOGFMIN))
            else:
                features.append(torch.log2(pitch))

        if emphases.PERIODICITY_FEATURE:
            features.append(periodicity)

        # Pitch and periodicity use floating-point hopsize, while mels and
        # loudness use an integer hopsize in samples. This results in
        # single-frame differences when the audio length is within one sample
        # of a new frame due to floating-point error. We simply repeat the last
        # frame in this rare case.
        frames = emphases.convert.samples_to_frames(audio.shape[-1])
        if pitch.shape[1] == frames + 1:
            pitch = pitch[:, :-1]
            periodicity = periodicity[:, :-1]

    # Preprocess loudness
    if emphases.LOUDNESS_FEATURE:
        loudness = emphases.data.preprocess.loudness.from_audio(
            audio,
            emphases.SAMPLE_RATE)
        features.append(loudness.to(audio.device))

    # Concatenate features
    features = features[0] if len(features) == 1 else torch.cat(features)

    return features[None]



import json

import numpy as np
import pypar
import torch
import torchaudio

import emphases


###############################################################################
# Dataset
###############################################################################


class Dataset(torch.utils.data.Dataset):

    def __init__(self, name, partition):
        self.cache = emphases.CACHE_DIR / name

        # Get list of stems
        with open(emphases.PARTITION_DIR / f'{name}.json') as file:
            self.stems = json.load(file)[partition]

        # Store lengths for bucketing
        audio_files = [
            self.cache / 'audio' / f'{stem}.wav' for stem in self.stems]
        self.lengths = [
            emphases.convert.samples_to_frames(
                torchaudio.info(audio_file).num_frames)
            for audio_file in audio_files]

        # Total number of frames
        self.frames = sum(self.lengths)

    def __getitem__(self, index):
        """Retrieve the indexth item"""
        stem = self.stems[index]

        # Load alignment
        alignment = pypar.Alignment(
            self.cache / 'alignment' / f'{stem}.TextGrid')

        # Compute word bounds
        bounds = alignment.word_bounds(
            emphases.SAMPLE_RATE,
            emphases.HOPSIZE,
            silences=True)
        word_bounds = torch.cat(
            [torch.tensor(bound)[None] for bound in bounds]).T

        # Load audio
        audio = emphases.load.audio(self.cache / 'audio' / f'{stem}.wav')

        features = []

        # Load mels
        if emphases.MEL_FEATURE:
            features.append(torch.load(self.cache / 'mels' / f'{stem}.pt'))

        # Load pitch
        if emphases.PITCH_FEATURE:
            pitch = torch.load(self.cache / 'pitch' / f'{stem}-pitch.pt')
            if emphases.NORMALIZE:
                features.append(
                    (torch.log2(pitch) - emphases.LOGFMIN) /
                    (emphases.LOGFMAX - emphases.LOGFMIN))
            else:
                features.append(torch.log2(pitch))

        # Load periodicity
        if emphases.PERIODICITY_FEATURE:
            periodicity = torch.load(
                self.cache / 'pitch' / f'{stem}-periodicity.pt')
            features.append(periodicity)

        # Load loudness
        if emphases.LOUDNESS_FEATURE:
            loudness = torch.load(self.cache / 'loudness' / f'{stem}.pt')
            features.append(loudness)

        # Concatenate
        features = features[0] if len(features) == 1 else torch.cat(features)

        # Load per-word ground truth emphasis scores
        scores = torch.load(self.cache / 'scores' / f'{stem}.pt')[None]

        return features, scores, word_bounds, alignment, audio, stem

    def __len__(self):
        """Length of the dataset"""
        return len(self.stems)

    def buckets(self):
        """Partition indices into buckets based on length for sampling"""
        # Get the size of a bucket
        size = len(self) // emphases.BUCKETS

        # Get indices in order of length
        indices = np.argsort(self.lengths)
        lengths = np.sort(self.lengths)

        # Split into buckets based on length
        buckets = [
            np.stack((indices[i:i + size], lengths[i:i + size])).T
            for i in range(0, len(self), size)]

        # Concatenate partial bucket
        if len(buckets) == emphases.BUCKETS + 1:
            residual = buckets.pop()
            buckets[-1] = np.concatenate((buckets[-1], residual), axis=0)

        return buckets



from scipy.signal import butter, lfilter


def butter_bandpass(lowcut, highcut, fs, order=5):
    """Generate the butter bandpass filter

    For more details see scipy.signal.butter documentation

    Parameters
    ----------
    lowcut: int
        The low cut value
    highcut: type
        description
    fs: int
        Signal sample rate
    order: int
        Order of the butter fiter

    Returns
    -------
    b: arraylike
    	Numerator polynomial of the IIR filter
    a: arraylike
    	Denominator polynomial of the IIR filter
    """
    nyq = .5 * fs
    low = lowcut / nyq
    if highcut >= nyq * .95:
        highcut = nyq * .95
    high = highcut / nyq
    b, a = butter(order, [low, high], btype='band')
    return b, a


def butter_bandpass_filter(data, lowcut, highcut, fs, order=5):
    """Filter signal data using a butter filter type

    For more details see scipy.signal.butter and scipy.signal.lfilter documentation

    Parameters
    ----------
    data: arraylike
        An N-dimensional input array.
    lowcut: int
        The lowcut filtering value.
    highcut: type
        The highcut filtering value.
    fs: int
        The signal sample rate.
    order: int
        The order of the butter filter.

    Returns
    -------
    arraylike
    	An N-dimensional filtered array
    """
    b, a = butter_bandpass(lowcut, highcut, fs, order=order)
    return lfilter(b, a, data)



import argparse
from pathlib import Path

import emphases


###############################################################################
# Scaling laws plot
###############################################################################


def parse_args():
    """Parse command-line arguments"""
    parser = argparse.ArgumentParser(
        description='Create scaling law figure')
    parser.add_argument(
        '--evaluations',
        type=str,
        nargs='+',
        required=True,
        help='The evaluations to plot')
    parser.add_argument(
        '--xlabel',
        type=str,
        required=True,
        help='Label for x axis')
    parser.add_argument(
        '--output_file',
        type=Path,
        required=True,
        help='The output jpg file')
    parser.add_argument(
        '--yticks',
        type=float,
        nargs='+',
        required=True,
        help='The y axis tick mark locations')
    parser.add_argument(
        '--sizes',
        type=int,
        nargs='+',
        help='The number of utterances used in each evaluation')
    parser.add_argument(
        '--scores',
        type=float,
        nargs='+',
        help='The Pearson Correlation y values')
    parser.add_argument(
        '--steps',
        type=int,
        nargs='+',
        help='The number of training steps')
    parser.add_argument(
        '--text_offsets',
        type=float,
        nargs='+',
        help='The amount to space the text below the plot point')
    return parser.parse_args()


if __name__ == '__main__':
    emphases.plot.scaling.scaling_laws(**vars(parse_args()))



import multiprocessing as mp
import warnings

import librosa
import numpy as np
import penn
import torch
import torchutil

import emphases


###############################################################################
# Interface
###############################################################################


def from_audio(audio, sample_rate=emphases.SAMPLE_RATE):
    """Compute mels from audio"""
    # Mayble resample
    audio = emphases.resample(audio, sample_rate)

    # Compute loudness
    return a_weighted(audio, sample_rate, hop_length=emphases.HOPSIZE)


def from_file(audio_file):
    """Load audio and compute mels"""
    audio = emphases.load.audio(audio_file)

    # Compute loudness
    return from_audio(audio)


def from_file_to_file(audio_file, output_file):
    """Compute loudness from audio file and save to disk"""
    loudness = from_file(audio_file)

    # Save to disk
    output_file.parent.mkdir(exist_ok=True, parents=True)
    torch.save(loudness, output_file)


def from_files_to_files(audio_files, output_files):
    """Compute loudness for many files and save to disk"""
    torchutil.multiprocess_iterator(
        wrapper,
        zip(audio_files, output_files),
        'Preprocessing a-weighted loudness',
        total=len(audio_files),
        num_workers=emphases.NUM_WORKERS)


###############################################################################
# Loudness
###############################################################################


def a_weighted(audio, sample_rate, hop_length=None, pad=False):
    """Retrieve the per-frame loudness"""
    # Save device
    device = audio.device

    # Default hop length of 10 ms
    hop_length = sample_rate // 100 if hop_length is None else hop_length

    if audio.dim() == 2:
        audio = audio[:, None, :]
    elif audio.dim() == 1:
        audio = audio[None, None, :]

    # Pad audio
    p = (emphases.NUM_FFT - emphases.HOPSIZE) // 2
    audio = torch.nn.functional.pad(audio, (p, p), "reflect").squeeze(1)

    # Convert to numpy
    audio = audio.detach().cpu().numpy().squeeze(0)

    # Cache weights
    if not hasattr(a_weighted, 'weights'):
        a_weighted.weights = perceptual_weights()

    # Take stft
    stft = librosa.stft(
        audio,
        n_fft=penn.WINDOW_SIZE,
        hop_length=hop_length,
        win_length=penn.WINDOW_SIZE,
        center=pad,
        pad_mode='constant')

    # Compute magnitude on db scale
    db = librosa.amplitude_to_db(np.abs(stft))

    # Apply A-weighting
    weighted = db + a_weighted.weights

    # Threshold
    weighted[weighted < emphases.MIN_DB] = emphases.MIN_DB

    # Average over weighted frequencies
    loudness = torch.from_numpy(weighted.mean(axis=0)).float().to(device)[None]

    # Scale to roughly [0, 1]
    if emphases.NORMALIZE:
        return (loudness + 100.) / 100.
    return loudness


def perceptual_weights():
    """A-weighted frequency-dependent perceptual loudness weights"""
    frequencies = librosa.fft_frequencies(
        sr=penn.SAMPLE_RATE,
        n_fft=penn.WINDOW_SIZE)

    # A warning is raised for nearly inaudible frequencies, but it ends up
    # defaulting to -100 db. That default is fine for our purposes.
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', RuntimeWarning)
        return librosa.A_weighting(frequencies)[:, None] - emphases.REF_DB

def wrapper(item):
    """Multiprocessing wrapper"""
    from_file_to_file(*item)



import numpy as np

import emphases


###############################################################################
# Constants
###############################################################################


SILENCE_SYMBOLS = [
    '#',
    '!pau',
    'sp',
    '<s>',
    'pau',
    '!sil',
    'sil',
    '',
    ' ',
    '<p>',
    '<p:>',
    '.',
    ',',
    '?',
    '<silent>']


###############################################################################
# Duration
###############################################################################


def _get_dur_stats(labels, rate=200):
    durations = []
    for i in range(len(labels)):
        st, en, unit = labels[i]
        st *= rate
        en *= rate
        if unit.lower() not in SILENCE_SYMBOLS:
            dur = en - st
            dur = np.log(dur + 1.)
            durations.append(dur)
    durations = np.array(durations)
    return np.min(durations), np.max(durations), np.mean(durations)


def get_rate(params, hp=10, lp=150):
    """
    estimation of speech rate as a center of gravity of wavelet spectrum
    similar to method described in "Boundary Detection using Continuous Wavelet Analysis" (2016)
    """
    params = emphases.baselines.prominence.smooth_and_interp.smooth(params, hp)
    params -= emphases.baselines.prominence.smooth_and_interp.smooth(params, lp)

    wavelet_matrix, *_ = emphases.baselines.prominence.cwt_utils.cwt_analysis(
        params,
        mother_name='Morlet',
        num_scales=80,
        scale_distance=.1,
        apply_coi=True,
        period=2)
    wavelet_matrix = abs(wavelet_matrix)

    rate = np.zeros(len(params))

    for i in range(0,wavelet_matrix.shape[1]):
        frame_en = np.sum(wavelet_matrix[:, i])
        # center of gravity
        rate[i] = np.nonzero(
            wavelet_matrix[:, i].cumsum() >= frame_en * .5)[0].min()

    return emphases.baselines.prominence.smooth_and_interp.smooth(rate, 30)


def duration(labels, rate=200):
    """Construct duration signal from labels"""
    dur = np.zeros(len(labels))
    params = np.zeros(int(labels[-1][1] * rate))
    prev_end = 0
    min_dur, *_ = _get_dur_stats(labels, rate=200)
    for i in range(0, len(labels)):
        st, en, unit = labels[i]
        st *= rate
        en *= rate
        dur[i] = en - st
        dur[i] = np.log(dur[i] + 1.)

        if unit.lower() in SILENCE_SYMBOLS:
            dur[i] = min_dur

        # skip very short units, likely labelling errors
        if en <= st + .01:
            continue

        # unit duration -> height of the duration contour in the middle of the unit
        index = min(len(params) - 1, int(st + (en - st) / 2.))
        params[index] = dur[i]

        # Handle gaps in labels similarly to silences
        if st > prev_end and i > 1:
            params[int(prev_end + (st - prev_end) / 2.)] = min_dur
        prev_end = en

    # set endpoints to mean in order to avoid large "valleys"
    params[0] = np.mean(dur)
    params[-1] = np.mean(dur)

    # make continous duration contour and smooth a bit
    params = emphases.baselines.prominence.smooth_and_interp.interpolate_zeros(params, 'pchip')
    return emphases.baselines.prominence.smooth_and_interp.smooth(params, 20)


def get_duration_signal(
    alignment,
    weights=[],
    rate=1):
    """
    Construct duration contour from labels. If many tiers are selected,
    construct contours for each tier and return a weighted sum of those
    """
    word_tier = [(word.start(), word.end(), str(word)) for word in alignment]
    phoneme_tier = [
        (phoneme.start(), phoneme.end(), str(phoneme))
    for phoneme in alignment.phonemes()]
    tiers = [phoneme_tier, word_tier]

    durations = []

    for tier in tiers:
        durations.append(
            emphases.baselines.prominence.normalize(
                duration(tier, rate=rate)))
    durations = match_length(durations)
    sum_durations = np.zeros(len(durations[0]))
    if len(weights) != len(tiers):
        weights = np.ones(len(tiers))
    for i in range(len(durations)):
        sum_durations += durations[i] * weights[i]
    return sum_durations


def match_length(sig_list):
    """Reduce length of all signals to a the minimum one.

    Parameters
    ----------
    sig_list: list
        List of signals which are 1D array of samples.
    """
    length = min(map(len, sig_list))
    for i in range(0, len(sig_list)):
        sig_list[i] = sig_list[i][:int(length)]
    return sig_list



import numpy as np
from operator import itemgetter


def simplify(loma):
    """?
    Parameters
    ----------
    loma: type
        description
    """
    simplified = []
    for l in loma:
        # align loma to it's position in the middle of the line
        pos =  l[int(len(l) / 2.)][0]
        strength = l[-1][1]
        simplified.append((pos, strength))
    return simplified


def get_prominences(pos_loma, alignment, rate=1):
    """?
    Parameters
    ----------
    pos_loma: list of ?
        Positive loma values
    labels: list of tuple (float, float, string)
        List of labels which are lists of 3 elements [start, end, description]
    """
    max_word_loma = []
    loma = simplify(pos_loma)
    for st, end in [(word.start(), word.end()) for word in alignment]:
        st *= rate
        end *= rate
        word_loma = []
        for l in loma:
            if l[0] >= st and l[0] <= end:
                word_loma.append(l)
        if len(word_loma) > 0:
            max_word_loma.append(sorted(word_loma, key=itemgetter(1))[-1])
        else:
            max_word_loma.append([st + (end - st) / 2., 0.])
    return max_word_loma


def get_boundaries(max_word_loma, boundary_loma, alignment):
    """get strongest lines of minimum amplitude between adjacent words' max lines"""
    boundary_loma = simplify(boundary_loma)
    max_boundary_loma = []
    st = 0
    end = 0
    for i in range(1, len(max_word_loma)):
        w_boundary_loma = []
        for l in boundary_loma:
            st = max_word_loma[i - 1][0]
            end = max_word_loma[i][0]
            if l[0] >= st and l[0] < end:
                if l[1] > 0:
                    w_boundary_loma.append(l)

        if len(w_boundary_loma) > 0:
            max_boundary_loma.append(
                sorted(w_boundary_loma, key=itemgetter(1))[-1])
        else:
            max_boundary_loma.append([st + (end - st) / 2, 0])

    # final boundary is not estimated
    max_boundary_loma.append((alignment.end(), 1))

    return max_boundary_loma


def _get_parent(child_index, parent_diff, parent_indices):
    """Private function to find the parent of the given child peak. At child peak index, follow the
    slope of parent scale upwards to find parent

    Parameters
    ----------
    child_index: int
        Index of the current child peak
    parent_diff: list of ?
        ?
    parent_indices: list of int ?
        Indices of available parents

    Returns
    _______
    int
    	The parent index or None if there is no parent
    """
    for i in range(0, len(parent_indices)):
        if parent_indices[i] > child_index:
            if parent_diff[int(child_index)] > 0:
                return parent_indices[i]
            else:
                if i > 0:
                    return parent_indices[i - 1]
                else:
                    return parent_indices[0]

    if len(parent_indices) > 0:
        return parent_indices[-1]


def get_loma(wavelet_matrix, scales, min_scale, max_scale):
    """Get the Line Of Maximum Amplitude (loma)

    Parameters
    ----------
    wavelet_matrix: matrix of float
        The wavelet matrix
    scales: list of int
        The list of scales
    min_scale: int
        The minimum scale
    max_scale: int
        The maximum scale

    Returns
    -------
    list of tuples
    	?

    Note
    ----
    change this so that one level is done in one chunk, not one parent.
    """
    min_peak = -10000. # minimum peak amplitude to consider. NOTE:this has no meaning unless scales normalized
    max_dist = 10 # how far in time to look for parent peaks. NOTE: frame rate and scale dependent

    # get peaks from the first scale
    peaks, indices = get_peaks(wavelet_matrix[min_scale], min_peak)

    loma = dict()
    root = dict()
    for i in range(0, len(peaks)):
        loma[indices[i]] = []

        # keep track of roots of each loma
        root[indices[i]] = indices[i]

    for i in range(min_scale + 1, max_scale):
        max_dist = np.sqrt(scales[i]) * 4

        # find peaks in the parent scale
        p_peaks, p_indices = get_peaks(wavelet_matrix[i], min_peak)
        parents = dict(zip(p_indices, p_peaks))

        # find a parent for each child peak
        children = dict()
        for p in p_indices:
            children[p] = []

        parent_diff = np.diff(wavelet_matrix[i], 1)
        for j in range(0, len(indices)):
            parent =_get_parent(indices[j], parent_diff, p_indices)
            if parent:
                if abs(parent - indices[j]) < max_dist and peaks[j] > min_peak:
                    children[parent].append([indices[j], peaks[j]])

        # for each parent, select max child
        peaks = []
        indices = []
        for p in children:
            if len(children[p]) > 0:
                maxi = sorted(children[p], key=itemgetter(1))[-1]
                indices.append(p)
                peaks.append(maxi[1] + parents[p])

                #append child to correct loma
                loma[root[maxi[0]]].append([maxi[0], maxi[1] + parents[p], i, p])
                root[p] = root[maxi[0]]

    sorted_loma = []
    for k in sorted(loma.keys()):
        if  len(loma[k]) > 0:
            sorted_loma.append(loma[k])

    return sorted_loma


def get_peaks(params, threshold=-10):
    """Find the peaks based on the given prosodic parameters.

    Parameters
    ----------
    params: ?
        Prosodic parameters
    threshold: int
        description

    Returns
    -------
    peaks: arraylike
        array of peak values and peak indices
    """
    indices = (np.diff(np.sign(np.diff(params))) < 0).nonzero()[0] + 1
    peaks = params[indices]
    return np.array([peaks[peaks > threshold], indices[peaks > threshold]])



import torch

import emphases


###############################################################################
# Sampler selection
###############################################################################


def sampler(dataset, partition):
    """Create batch sampler"""
    # Deterministic random sampler for training
    if partition in ['train', 'valid']:
        return Sampler(dataset)

    # Sample validation and test data sequentially
    elif partition.startswith('test'):
        return torch.utils.data.BatchSampler(
            torch.utils.data.SequentialSampler(dataset),
            1,
            False)

    else:
        raise ValueError(f'Partition {partition} is not defined')


###############################################################################
# Samplers
###############################################################################


class Sampler:

    def __init__(self, dataset, max_frames=emphases.MAX_TRAINING_FRAMES):
        self.max_frames = max_frames
        self.epoch = 0
        self.length = len(dataset)
        self.buckets = dataset.buckets()

    def __iter__(self):
        return iter(self.batch())

    def __len__(self):
        return len(self.batch())

    def batch(self):
        """Produces batch indices for one epoch"""
        # Deterministic shuffling based on epoch
        generator = torch.Generator()
        generator.manual_seed(emphases.RANDOM_SEED + self.epoch)

        # Iterate over length-partitioned buckets
        batches = []
        for bucket in self.buckets:

            # Shuffle bucket
            bucket = bucket[
                torch.randperm(len(bucket), generator=generator).tolist()]

            # Variable batch size
            batch = []
            max_length = 0
            for index, length in bucket:
                max_length = max(max_length, length)
                if (
                    batch and
                    (len(batch) + 1) * max_length > self.max_frames
                ):
                    batches.append(batch)
                    max_length = length
                    batch = [index]
                else:
                    batch.append(index)

        # Shuffle
        return [
            batches[i] for i in
            torch.randperm(len(batches), generator=generator).tolist()]

    def set_epoch(self, epoch):
        self.epoch = epoch



import contextlib
import functools
import os
import tempfile
from pathlib import Path
from typing import List, Optional, Tuple, Type, Union

import huggingface_hub
import pyfoal
import pypar
import torch
import torchutil
import torchaudio

import emphases


###############################################################################
# Emphasis annotation API
###############################################################################


def from_file(
    text_file: Union[str, bytes, os.PathLike],
    audio_file: Union[str, bytes, os.PathLike],
    checkpoint: Optional[Union[str, bytes, os.PathLike]] = None,
    batch_size: Optional[int] = None,
    gpu: Optional[int] = None
) -> Tuple[Type[pypar.Alignment], torch.Tensor]:
    """Produce emphasis scores for each word for files on disk

    Args:
        text_file: The speech transcript (.txt) or alignment (.TextGrid) file
        audio_file: The speech waveform audio file
        checkpoint: The model checkpoint to use for inference
        batch_size: The maximum number of frames per batch
        gpu: The index of the gpu to run inference on

    Returns:
        alignment: The forced phoneme alignment
        scores: The float-valued emphasis scores for each word
    """
    # Load audio
    audio = emphases.load.audio(audio_file)

    if str(text_file).endswith('.TextGrid'):

        # Load alignment
        alignment = pypar.Alignment(text_file)

        # Detect emphases
        return from_alignment_and_audio(
            alignment,
            audio,
            emphases.SAMPLE_RATE,
            checkpoint,
            batch_size,
            gpu)

    else:

        # Load text
        with open(text_file, encoding='utf-8') as file:
            text = file.read()

        # Detect emphases
        return from_text_and_audio(
            text,
            audio,
            emphases.SAMPLE_RATE,
            checkpoint,
            batch_size,
            gpu)


def from_file_to_file(
    text_file: List[Union[str, bytes, os.PathLike]],
    audio_file: List[Union[str, bytes, os.PathLike]],
    output_prefix: Optional[List[Union[str, bytes, os.PathLike]]] = None,
    checkpoint: Optional[Union[str, bytes, os.PathLike]] = None,
    batch_size: Optional[int] = None,
    gpu: Optional[int] = None
) -> None:
    """Produce emphasis scores for each word for files on disk and save to disk

    Args:
        text_file: The speech transcript (.txt) or alignment (.TextGrid) file
        audio_file: The speech waveform audio file
        output_prefix: The output prefix. Defaults to text file stem.
        checkpoint: The model checkpoint to use for inference
        batch_size: The maximum number of frames per batch
        gpu: The index of the gpu to run inference on
    """
    if output_prefix is None:
        output_prefix = text_file.stem

    # Detect emphases
    results = from_file(
        text_file,
        audio_file,
        checkpoint,
        batch_size,
        gpu)
    if text_file.name.endswith('.txt'):
        alignment, prominence = results
    else:
        alignment = pypar.Alignment(text_file)
        prominence = results

    # Save results
    alignment.save(f'{output_prefix}.TextGrid')
    torch.save(results.cpu(), f'{output_prefix}.pt')


def from_files_to_files(
    text_files: List[Union[str, bytes, os.PathLike]],
    audio_files: List[Union[str, bytes, os.PathLike]],
    output_prefixes: Optional[List[Union[str, bytes, os.PathLike]]] = None,
    checkpoint: Optional[Union[str, bytes, os.PathLike]] = None,
    batch_size: Optional[int] = None,
    gpu: Optional[int] = None
) -> None:
    """Produce emphasis scores for each word for many files and save to disk

    Args:
        text_file: The speech transcript (.txt) or alignment (.TextGrid) files
        audio_files: The corresponding speech audio files
        output_prefixes: The output files. Defaults to text file stems.
        checkpoint: The model checkpoint to use for inference
        batch_size: The maximum number of frames per batch
        gpu: The index of the gpu to run inference on
    """
    # Set default output path
    if output_prefixes is None:
        output_prefixes = [file.stem for file in text_files]

    # Batch forced alignments to improve performance via multiprocessing
    with tempfile.TemporaryDirectory() as directory:
        directory = Path(directory)

        # Get files to force-align
        try:
            text_filtered, audio_filtered = zip(*[
                (text, audio) for text, audio in zip(text_files, audio_files)
                if str(text).endswith('.txt')])

            # Get location to save files
            output_filtered = [
                directory / Path(file).with_suffix('.TextGrid')
                for file in text_filtered]

            # Force align
            pyfoal.from_files_to_files(
                text_filtered,
                audio_filtered,
                output_filtered,
                aligner='p2fa')

        except ValueError:
            pass

        # Update filenames for emphasis detection
        text_files, audio_files = zip(*[
            (text, audio) if str(text).endswith('TextGrid')
            else (directory / Path(text).with_suffix('.TextGrid'), audio)
            for text, audio in zip(text_files, audio_files)])

        # Detect emphases
        annotation_fn = functools.partial(
            from_file_to_file,
            checkpoint=checkpoint,
            batch_size=batch_size,
            gpu=gpu)
        for files in torchutil.iterator(
            zip(text_files, audio_files, output_prefixes),
            emphases.CONFIG,
            total=len(text_files)
        ):
            annotation_fn(*files)


def from_text_and_audio(
    text: str,
    audio: torch.Tensor,
    sample_rate: int,
    checkpoint: Optional[Union[str, bytes, os.PathLike]] = None,
    batch_size: Optional[int] = None,
    gpu: Optional[int] = None
) -> Tuple[Type[pypar.Alignment], torch.Tensor]:
    """Produce emphasis scores for each word

    Args:
        text: The speech transcript
        audio: The speech waveform
        sample_rate: The audio sampling rate
        checkpoint: The model checkpoint to use for inference
        batch_size: The maximum number of frames per batch
        gpu: The index of the gpu to run inference on

    Returns:
        alignment: The forced phoneme alignment
        scores: The float-valued emphasis scores for each word
    """
    # Get word alignment
    alignment = pyfoal.from_text_and_audio(
        text,
        audio,
        sample_rate,
        aligner='p2fa')

    # Infer
    scores = from_alignment_and_audio(
        alignment,
        audio,
        sample_rate,
        checkpoint,
        batch_size,
        gpu)

    return alignment, scores


def from_alignment_and_audio(
    alignment: pypar.Alignment,
    audio: torch.Tensor,
    sample_rate: int,
    checkpoint: Optional[Union[str, bytes, os.PathLike]] = None,
    batch_size: Optional[int] = None,
    gpu: Optional[int] = None
) -> Tuple[Type[pypar.Alignment], torch.Tensor]:
    """Produce emphasis scores for each word

    Args:
        alignment: The forced phoneme alignment
        audio: The speech waveform
        sample_rate: The audio sampling rate
        checkpoint: The model checkpoint to use for inference
        batch_size: The maximum number of frames per batch
        gpu: The index of the gpu to run inference on

    Returns:
        scores: The float-valued emphasis scores for each word
    """
    # Neural method
    if emphases.METHOD == 'neural':

        scores = []

        # Preprocess audio
        for features, word_bounds in preprocess(
            alignment,
            audio,
            sample_rate,
            batch_size,
            gpu
        ):

            # Infer
            logits = infer(features, word_bounds, checkpoint).detach()[0]

            # Postprocess
            scores.append(postprocess(logits))

        # Concatenate results
        return torch.cat(scores, 1)

    # Prominence method
    if emphases.METHOD == 'prominence':
        return emphases.baselines.prominence.infer(
            alignment,
            audio,
            sample_rate)

    # Pitch variance method
    if emphases.METHOD == 'pitch-variance':
        return emphases.baselines.pitch_variance.infer(
            alignment,
            audio,
            sample_rate,
            gpu)

    # Duration variance method
    if emphases.METHOD == 'duration-variance':
        return emphases.baselines.duration_variance.infer(alignment)

    raise ValueError(
        f'Emphasis annotation method {emphases.METHOD} is not defined')


###############################################################################
# Inference steps
###############################################################################


def infer(features, word_bounds, checkpoint=None):
    """Perform model inference to annotate emphases of each word"""
    # Maybe cache model
    if (
        not hasattr(infer, 'model') or
        infer.checkpoint != checkpoint or
        infer.device_type != features.device.type
    ):
        # Initialize model
        model = emphases.Model()

        # Maybe download from HuggingFace
        if checkpoint is None:
            checkpoint = huggingface_hub.hf_hub_download(
                'maxrmorrison/emphases',
                'model.pt')

        # Load from disk
        infer.model, *_ = torchutil.checkpoint.load(checkpoint, model)
        infer.checkpoint = checkpoint
        infer.device_type = features.device.type

    # Move model to correct device (no-op if devices are the same)
    infer.model = infer.model.to(features.device)

    # Use full sequence lengths
    frame_lengths = torch.tensor(
        [features.shape[-1]],
        dtype=torch.long,
        device=features.device)
    word_lengths = torch.tensor(
        [word_bounds.shape[-1]],
        dtype=torch.long,
        device=features.device)

    # Infer
    with emphases.inference_context(infer.model):
        return infer.model(features, frame_lengths, word_bounds, word_lengths)


def postprocess(logits):
    """Postprocess network output"""
    if emphases.METHOD == 'neural':
        if emphases.LOSS == 'bce':
            return torch.sigmoid(logits)
        elif emphases.LOSS == 'mse':
            return torch.clamp(logits, 0., 1.)
    return logits


def preprocess(
    alignment,
    audio,
    sample_rate=emphases.SAMPLE_RATE,
    batch_size=None,
    gpu=None):
    """Convert audio to model input"""
    # Resample
    if sample_rate != emphases.SAMPLE_RATE:
        audio = resample(audio, sample_rate)

    # Pad audio and get total number of frames
    padding = int((emphases.WINDOW_SIZE - emphases.HOPSIZE) / 2)
    audio = torch.nn.functional.pad(audio, (padding, padding))
    total_frames = int(audio.shape[-1] / emphases.HOPSIZE)

    # Default to running all frames in a single batch
    batch_size = total_frames if batch_size is None else batch_size

    # Generate batches
    start = 0
    while start < len(alignment):

        # Accumulate enough frames for this batch
        frames = 0.
        end = start + 1
        while end < len(alignment):

            # Update frames
            frames += emphases.convert.seconds_to_frames(
                alignment[end - 1].duration())

            # Stop if we've accumulated enough frames
            if int(frames) > batch_size:
                break

            end += 1

        # Slice alignment
        batch_alignment = alignment[start:end]

        # Compute word bounds
        bounds = batch_alignment.word_bounds(
            emphases.SAMPLE_RATE,
            emphases.HOPSIZE,
            silences=True)
        batch_word_bounds = torch.cat(
            [torch.tensor(bound)[None] for bound in bounds]).T[None]

        # Slice audio at frame boundaries
        start_sample = int(emphases.convert.frames_to_samples(
            int(emphases.convert.seconds_to_frames(
                alignment[start].start()))))
        end_sample = int(emphases.convert.frames_to_samples(
            int(emphases.convert.seconds_to_frames(
                alignment[end - 1].end()))))
        batch_audio = audio[:, start_sample:end_sample]

        try:

            # Preprocess audio
            batch_features = emphases.data.preprocess.from_audio(
                batch_audio,
                gpu)

            # Run inference
            yield batch_features, batch_word_bounds

        # Handle residual frame
        except RuntimeError:
            pass

        # Update start word
        start = end


###############################################################################
# Word and frame resolution resampling
###############################################################################


def downsample(xs, word_bounds, word_lengths):
    """Interpolate from frame to word resolution"""
    # Average resampling
    if emphases.DOWNSAMPLE_METHOD in ['average', 'max', 'sum']:

        # Allocate memory for word resolution sequence
        result = torch.zeros(
            (xs.shape[0], xs.shape[1], word_lengths.max().item()),
            dtype=xs.dtype,
            device=xs.device)

        # Iterate over batch
        for i, (x, bounds, length) in enumerate(
            zip(xs, word_bounds, word_lengths)
        ):

            # Iterate over words
            for j in range(length):

                # Get word bounds
                start, end = bounds[0, j], bounds[1, j]

                # Downsample
                if emphases.DOWNSAMPLE_METHOD == 'average':
                    result[i, :, j] = x[:, start:end].mean(dim=1)
                elif emphases.DOWNSAMPLE_METHOD == 'max':
                    result[i, :, j] = x[:, start:end].max(dim=1).values
                elif emphases.DOWNSAMPLE_METHOD == 'sum':
                    result[i, :, j] = x[:, start:end].sum(dim=1)

        return result

    # Centerpoint resampling
    if emphases.DOWNSAMPLE_METHOD == 'center':

        # Downsample to words via index selection
        indices = (word_bounds[:, 0] + word_bounds[:, 1]) // 2
        return xs.transpose(1, 2)[
            torch.arange(xs.shape[0])[:, None],
            indices
        ].transpose(1, 2)

    raise ValueError(
        f'Interpolation method {emphases.DOWNSAMPLE_METHOD} is not defined')


def upsample(xs, word_bounds, word_lengths, frame_lengths):
    """Interpolate from word to frame resolution"""
    result = torch.zeros(
        (xs.shape[0], xs.shape[1], frame_lengths.max()),
        dtype=xs.dtype,
        device=xs.device)
    for i, (x, word_bound, word_length, frame_length) in enumerate(
        zip(xs, word_bounds, word_lengths, frame_lengths)
    ):

        # Truncate to valid sequence
        x = x[..., :word_length]
        word_bound = word_bound[..., :word_length]

        # Get center time of each word in frames
        word_times = (
            word_bound[0] + (word_bound[1] - word_bound[0]) / 2.)[None]

        # Get frame centers
        frame_times = .5 + torch.arange(frame_length, device=xs.device)[None]

        # Single value edge case
        if x.shape[1] == 1:
            result[i, :, :frame_length] = x[0]

        # Linear interpolation
        elif emphases.UPSAMPLE_METHOD == 'linear':

            # Compute slope and intercept at original times
            slope = (
                (x[:, 1:] - x[:, :-1]) /
                (word_times[:, 1:] - word_times[:, :-1]))
            intercept = x[:, :-1] - slope.mul(word_times[:, :-1])

            # Compute indices at which we evaluate points
            indices = torch.sum(
                torch.ge(frame_times[:, :, None], word_times[:, None, :]),
                -1
            ) - 1
            indices = torch.clamp(indices, 0, slope.shape[-1] - 1)

            # Compute index into parameters
            line_idx = torch.linspace(
                0,
                indices.shape[0],
                1,
                device=indices.device).to(torch.long)
            line_idx = line_idx.expand(indices.shape)

            # Interpolate
            result[i, :, :frame_length] = (
                slope[line_idx, indices].mul(frame_times) +
                intercept[line_idx, indices])

        # Nearest neighbors interpolation
        elif emphases.UPSAMPLE_METHOD == 'nearest':

            # Compute indices at which we evaluate points
            indices = torch.sum(
                torch.ge(frame_times[:, :, None], word_times[:, None, :]),
                -1
            ) - 1
            indices = torch.clamp(indices, 0, word_times.shape[-1] - 1)

            # Get nearest score
            result[i, :, :frame_length] = torch.index_select(x, 1, indices[0])

        else:
            raise ValueError(
                f'Interpolation method {emphases.UPSAMPLE_METHOD} ' +
                'is not defined')

    return result


###############################################################################
# Word segmentation
###############################################################################


def segment(xs, word_bounds, word_lengths):
    """Convert acoustic features to word segments"""
    # Get maximum word length
    max_length = (word_bounds[:, 1] - word_bounds[:, 0]).max()

    # Allocate memory
    batch = word_bounds.shape[0] * word_bounds.shape[2]
    result = torch.zeros(
        (batch, xs.shape[1], max_length),
        dtype=xs.dtype,
        device=xs.device)
    result_bounds = torch.zeros(
        (batch, 2, 1),
        dtype=torch.long,
        device=xs.device)
    result_lengths = torch.zeros((batch,), dtype=torch.long, device=xs.device)

    # Iterate over batch
    for i, (x, bounds, words) in enumerate(zip(xs, word_bounds, word_lengths)):

        # Iterate over words
        for j in range(bounds.shape[1]):
            k = min(j, words - 1)

            # Get word bounds
            start, end = bounds[0, k], bounds[1, k]

            # Chunk
            frames = end - start
            index = i * bounds.shape[1] + j
            result[index, :, :frames] = x[:, start:end]
            result_bounds[index, 1, 0] = frames
            result_lengths[index] = frames

    return result, result_bounds, result_lengths


###############################################################################
# Utilities
###############################################################################


@contextlib.contextmanager
def inference_context(model):
    """Prepare model for inference"""
    device_type = next(model.parameters()).device.type

    # Prepare model for evaluation
    model.eval()

    # Turn off gradient computation
    with torch.no_grad():

        # Automatic mixed precision on GPU
        with torch.autocast(device_type):
            yield

    # Prepare model for training
    model.train()


def resample(audio, sample_rate, target_rate=emphases.SAMPLE_RATE):
    """Perform audio resampling"""
    if sample_rate == target_rate:
        return audio
    resampler = torchaudio.transforms.Resample(sample_rate, target_rate)
    resampler = resampler.to(audio.device)
    return resampler(audio)



import numpy as np

import scipy.signal

import emphases


def _get_f0(spec, energy, min_hz, max_hz, thresh, sil_thresh):
    """
    return frequency bin with maximum energy, if it is over given threshold
    and overall energy of the frame is over silence threshsold
    otherwise return 0 (unvoiced)
    """
    cand = int(min_hz) + np.argmax(spec[int(min_hz):int(max_hz)])
    if spec[cand] > thresh and energy > sil_thresh:
        if cand > 2 * min_hz and spec[int(round(cand / 2.))] > spec[cand] * .5:
            return int(round(cand / 2.))
        else:
            return cand
    return 0


def _track_pitch(
    pic,
    min_hz=50,
    max_hz=450,
    thresh=.1,
    energy_thresh=1.):
    """
    extract pitch contour from time-frequency image
    bin with maximum energy / frame is chosen as a first f0 estimate,
    following with refinement steps based on the assumption of continuity of the pitch track
    """
    pitch = np.zeros(pic.shape[0])

    # calc energy threshold for voicing
    log_energy = np.log(np.sum(pic, axis=1))
    energy_thresh = \
        np.min(emphases.baselines.prominence.smooth_and_interp.smooth(log_energy, 20)) + energy_thresh
    pic_smooth = pic * scipy.ndimage.gaussian_filter(pic, [2, 5])

    # find frequency bins with max_energy
    for i in range(0, pic_smooth.shape[0]):
        pitch[i] = _get_f0(
            pic_smooth[i],
            log_energy[i],
            min_hz,
            max_hz,
            thresh,
            energy_thresh)

    # second pass with soft constraints
    n_iters = 3
    from scipy.signal import gaussian

    for iter in range(0, n_iters):
        smoothed = emphases.baselines.prominence.f0_processing.process(pitch)
        smoothed = emphases.baselines.prominence.smooth_and_interp.smooth(smoothed, int(200. / (iter + 1.)))

        # gradually thightening gaussian window centered on current estimate to softly constrain next iteration
        win_len = 800
        g_window = gaussian(win_len, int(np.mean(smoothed) * (1. / (iter + 1.) ** 2)))

        for i in range(0, pic.shape[0]):
            window = np.zeros(len(pic_smooth[i]))
            st = int(np.max((0, int(smoothed[i] - win_len))))
            end = int(np.min((int(smoothed[i] + win_len * .5), win_len - st)))
            window[st:end] = g_window[win_len - end:]
            pitch[i] = _get_f0(
                pic_smooth[i] * window, log_energy[i],
                min_hz,
                max_hz,
                thresh,
                energy_thresh)

    return pitch


def _assign_to_bins(pic, freqs, mags):
    for i in range(1, freqs.shape[0] - 1):
        for j in range(0, freqs.shape[1]):
            try:
                pic[j, int(freqs[i, j])] += mags[i, j]
            except:
                pass


def inst_freq_pitch(
    wav_form,
    fs,
    min_hz=emphases.FMIN,
    max_hz=emphases.FMAX,
    voicing_thresh=emphases.VOICED_THRESHOLD,
    target_rate=200):
    """Extract speech f0 using the continuous wavelet transform"""
    voicing_thresh = (voicing_thresh - 50.) / 100.
    sample_rate = 4000
    tmp_wav_form = emphases.baselines.prominence.resample(wav_form, fs, sample_rate)
    tmp_wav_form = emphases.baselines.prominence.normalize(tmp_wav_form)

    DEC = int(round(sample_rate / target_rate))

    pic = np.zeros(
        shape=(int(len(tmp_wav_form) / float(DEC)), int(sample_rate / 4.)))

    # use continuous wavelet transform to get instantenous frequencies
    # integrate analyses with morlet mother wavelets with period = 5 for
    # good time and frequency resolution
    # setup wavelet
    s0 = 2. / sample_rate
    dj = .05 # 20 scales per octave
    J = 120  # six octaves
    dt = 1. / sample_rate
    periods = [5]
    for p in periods:
        wavelet_matrix, *_ = emphases.baselines.prominence.cwt_utils.cwt_analysis(
            tmp_wav_form,
            mother_name='morlet',
            first_scale=s0,
            num_scales=J,
            scale_distance=dj,
            apply_coi=False,
            period=p,
            frame_rate=sample_rate)

        # hilbert transform
        phase = np.unwrap(np.angle(wavelet_matrix), axis=1)
        freqs =  np.abs((np.gradient(phase, dt)[1]) / (2. * np.pi))

        freqs = scipy.signal.decimate(freqs, DEC, zero_phase=True)
        mags = scipy.signal.decimate(abs(wavelet_matrix), DEC, zero_phase=True)

        # normalize magnitudes
        mags = (mags - mags.min()) / mags.ptp()

        # construct time-frequency image
        _assign_to_bins(pic, freqs, mags)

    # perform frequency domain autocorrelation to enhance f0
    pic = scipy.ndimage.filters.gaussian_filter(pic, [1, 1])
    length = np.min((max_hz * 3, pic.shape[1])).astype(int)

    for i in range(0, pic.shape[0]):
        acorr1 = np.correlate(pic[i, :length], pic[i, :length], mode='same')
        pic[i, :int(length / 2.)] *= acorr1[int(len(acorr1) / 2.):]

    return _track_pitch(pic, min_hz, max_hz, voicing_thresh)



import argparse
from pathlib import Path

import emphases


###############################################################################
# Determine which words in a speech file are emphasized
###############################################################################


def parse_args():
    """Parse command-line arguments"""
    parser = argparse.ArgumentParser(
        description='Determine which words in a speech file are emphasized')
    parser.add_argument(
        '--text_files',
        type=Path,
        nargs='+',
        required=True,
        help='The speech transcript (.txt) or alignment (.TextGrid) files')
    parser.add_argument(
        '--audio_files',
        type=Path,
        nargs='+',
        required=True,
        help='The corresponding speech audio files')
    parser.add_argument(
        '--output_prefixes',
        type=Path,
        nargs='+',
        required=False,
        help='output_prefixes: The output files. Defaults to text files stems.')
    parser.add_argument(
        '--checkpoint',
        type=Path,
        help='The model checkpoint to use for inference')
    parser.add_argument(
        '--batch_size',
        type=int,
        help='The maximum number of frames per batch')
    parser.add_argument(
        '--gpu',
        type=int,
        help='The index of the gpu to run inference on')
    return parser.parse_args()


if __name__ == '__main__':
    emphases.from_files_to_files(**vars(parse_args()))



import penn
import torch

import emphases


###############################################################################
# Pitch variance method
###############################################################################


def infer(alignment, audio, sample_rate, gpu=None):
    """Compute per-word emphasis scores using pitch variance method"""
    # Infer pitch and periodicity
    pitch, _ = penn.from_audio(
        audio,
        sample_rate,
        hopsize=emphases.HOPSIZE_SECONDS,
        fmin=emphases.FMIN,
        fmax=emphases.FMAX,
        pad=True,
        interp_unvoiced_at=emphases.VOICED_THRESHOLD,
        gpu=gpu)

    # Compute pitch statistics in base-two log-space
    pitch = torch.log2(pitch)

    # Compute utterance statistics
    utterance_spread = spread(pitch)

    # Compute word statistics
    word_spreads = []
    for word in alignment:
        start = int(emphases.convert.seconds_to_frames(word.start()))
        end = int(emphases.convert.seconds_to_frames(word.end()))
        word_spreads.append(spread(pitch[0, start:end]))
    word_spreads = torch.tensor(
        word_spreads,
        dtype=pitch.dtype,
        device=pitch.device)[None]

    # Zero-center
    return word_spreads - utterance_spread


###############################################################################
# Utilities
###############################################################################


def spread(pitch):
    """Compute pitch spread"""
    return torch.quantile(pitch, .95) - torch.quantile(pitch, .05)



import argparse
from pathlib import Path

import emphases


###############################################################################
# Annotate emphases
###############################################################################


def parse_args():
    """Parse command-line arguments"""
    parser = argparse.ArgumentParser(description='Perform emphasis annotation')
    parser.add_argument(
        '--annotation_config',
        type=Path,
        default=emphases.DEFAULT_ANNOTATION_CONFIG,
        help='The ReSEval configuration file for the annotation task')
    parser.add_argument(
        '--dataset',
        default='libritts',
        help='The dataset to annotate')
    parser.add_argument(
        '--directory',
        type=Path,
        default=emphases.ANNOTATION_DIR,
        help='The directory to save results to')
    parser.add_argument(
        '--remote',
        action='store_true',
        help='Run subjective evaluation remotely')
    parser.add_argument(
        '--production',
        action='store_true',
        help='Deploy the subjective evaluation to crowdsource participants')
    parser.add_argument(
        '--interval',
        type=int,
        default=120,
        help='The time between monitoring updates in seconds')
    return parser.parse_args()


if __name__ == '__main__':
    emphases.annotate.datasets(**vars(parse_args()))



import numpy as np
from scipy import interpolate


def remove_bias(params, win_len=300):
    return params - smooth(params, win_len)


def interpolate_zeros(params, method='pchip', min_val=0):
    """
    Interpolate 0 values
    :param params: 1D data vector
    :param method:
    :param factor: factor for interpolation (must be integer)
    :return: interpolated 1D vector by a given factor
    """
    voiced = np.array(params, float)
    for i in range(0, len(voiced)):
        if voiced[i] == min_val:
            voiced[i] = np.nan

    if np.isnan(voiced[-1]):
        voiced[-1] = np.nanmin(voiced)
    if np.isnan(voiced[0]):
        voiced[0] = np.nanmean(voiced)

    not_nan = np.logical_not(np.isnan(voiced))

    indices = np.arange(len(voiced))
    if method == 'spline':
        interp = interpolate.UnivariateSpline(
            indices[not_nan],
            voiced[not_nan],
            k=2,
            s=0)
        # return voiced parts intact
        smoothed = interp(indices)
        for i in range(0, len(smoothed)):
            if not np.isnan(voiced[i]):
                smoothed[i] = params[i]
        return smoothed

    elif method == 'pchip':
        interp = interpolate.pchip(indices[not_nan], voiced[not_nan])
    else:
        interp = interpolate.interp1d(
            indices[not_nan],
            voiced[not_nan],
            method)
    return interp(indices)


def smooth(params, win, type='HAMMING'):
    """gaussian type smoothing, convolution with hamming window"""
    win = int(win + .5)
    if win >= len(params) - 1:
        win = len(params) - 1

    if win % 2 == 0:
        win += 1

    s = np.r_[params[win - 1:0:-1], params, params[-1:-win:-1]]

    if type == 'HAMMING':
        w = np.hamming(win)
    else:
        w = np.ones(win)

    y = np.convolve(w / w.sum(), s, mode='valid')
    return y[int(win / 2):-int(win / 2)]


def peak_smooth(params, max_iter, win, min_win=2, voicing=[]):
    """Iterative smoothing while preserving peaks, 'true envelope' -style"""
    smoothed = np.array(params)
    win_reduce = np.exp(np.linspace(np.log(win), np.log(min_win), max_iter))

    for i in range(0, max_iter):

        smoothed = np.maximum(params, smoothed)

        if len(voicing) > 0:
            smoothed = smooth(smoothed, int(win + .5))
            smoothed[voicing > 0] = params[voicing > 0]
        else:
            smoothed = smooth(smoothed, int(win + .5), type='rectangle')

        win = win_reduce[i]

    return smoothed



from setuptools import find_packages, setup


with open('README.md', encoding='utf8') as file:
    long_description = file.read()


setup(
    name='emphases',
    description='Crowdsourced and Automatic Speech Prominence Estimation',
    version='0.0.1',
    author='Interactive Audio Lab',
    author_email='interactiveaudiolab@gmail.com',
    url='https://github.com/interactiveaudiolab/emphases',
    install_requires=[
        'huggingface-hub',
        'librosa',
        'matplotlib',
        'numpy',
        'penn',
        'pycwt',
        'pyfoal',
        'pypar',
        'pyyaml',
        'reseval',
        'scipy',
        'torch',
        'torchutil',
        'torchaudio',
        'yapecs'],
    packages=find_packages(),
    package_data={'emphases': ['assets/*', 'assets/*/*']},
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords=['annotatation', 'audio', 'emphasis', 'prominence', 'speech'],
    classifiers=['License :: OSI Approved :: MIT License'],
    license='MIT')



import numpy as np

import emphases


def extract_energy(
    waveform,
    sample_rate=16000,
    min_freq=emphases.PROMINENCE_ENERGY_MIN,
    max_freq=emphases.PROMINENCE_ENERGY_MAX,
    frame_rate=200):
    # Get butterworth bandpass filter parameters
    lp_waveform =  emphases.baselines.prominence.filter.butter_bandpass_filter(
        waveform,
        min_freq,
        max_freq,
        sample_rate,
        order=5)

    # Compute energy
    energy = np.sqrt(lp_waveform ** 2)

    # Resample to frame rate
    return emphases.baselines.prominence.resample(energy, sample_rate, frame_rate)



import argparse

import emphases


###############################################################################
# Entry point
###############################################################################


def parse_args():
    """Parse command-line arguments"""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--datasets',
        nargs='+',
        default=emphases.DATASETS,
        help='The datasets to preprocess')
    parser.add_argument(
        '--gpu',
        type=int,
        help='The index of the gpu to run inference on')
    return parser.parse_known_args()[0]


if __name__ == '__main__':
    emphases.data.preprocess.datasets(**vars(parse_args()))



import torch

import emphases


###############################################################################
# Model definition
###############################################################################


class Model(torch.nn.Module):

    def __init__(self):
        super().__init__()

        # Input projection
        self.input_layer = torch.nn.Conv1d(
            emphases.NUM_FEATURES,
            emphases.CHANNELS,
            kernel_size=emphases.ENCODER_KERNEL_SIZE,
            padding='same')

        # Frame encoder
        self.frame_encoder = emphases.model.Layers(
            kernel_size=emphases.ENCODER_KERNEL_SIZE)

        # If we are resampling within the model, initialize word decoder
        if emphases.DOWNSAMPLE_LOCATION in ['input', 'intermediate']:
            self.word_decoder = emphases.model.Layers(
            kernel_size=emphases.DECODER_KERNEL_SIZE)

        # Output projection
        self.output_layer = torch.nn.Conv1d(
            emphases.CHANNELS,
            1,
            kernel_size=emphases.DECODER_KERNEL_SIZE,
            padding='same')

    def forward(self, features, frame_lengths, word_bounds, word_lengths):

        if emphases.DOWNSAMPLE_LOCATION == 'input':

            # Segment acoustic features into word segments
            segments, bounds, lengths = emphases.segment(
                features,
                word_bounds,
                word_lengths)

            # Embed frames
            frame_embeddings = self.frame_encoder(
                self.input_layer(segments),
                lengths)

            # Downsample
            if emphases.DOWNSAMPLE_METHOD == 'average':
                word_embeddings = frame_embeddings.mean(dim=2, keepdim=True)
            elif emphases.DOWNSAMPLE_METHOD == 'max':
                word_embeddings = frame_embeddings.max(
                    dim=2,
                    keepdim=True
                ).values
            elif emphases.DOWNSAMPLE_METHOD == 'sum':
                word_embeddings = frame_embeddings.sum(dim=2, keepdim=True)
            elif emphases.DOWNSAMPLE_METHOD == 'center':
                word_embeddings = emphases.downsample(
                    frame_embeddings,
                    bounds,
                    torch.ones(
                        (len(lengths),),
                        dtype=torch.long,
                        device=lengths.device))
            else:
                raise ValueError(
                    f'Interpolation method {emphases.DOWNSAMPLE_METHOD} is not defined')

            # Stitch together word segment embeddings
            mask = mask_from_lengths(word_lengths)
            word_embeddings = word_embeddings.squeeze(2).transpose(0, 1).reshape(
                word_embeddings.shape[1],
                word_bounds.shape[0],
                word_bounds.shape[2]
            ).permute(1, 0, 2) * mask

            # Decode
            word_embeddings = self.word_decoder(
                word_embeddings,
                word_lengths)

        else:

            # Embed frames
            frame_embeddings = self.frame_encoder(
                self.input_layer(features),
                frame_lengths)

            if emphases.DOWNSAMPLE_LOCATION == 'intermediate':

                # Downsample activations to word resolution
                word_embeddings = emphases.downsample(
                    frame_embeddings,
                    word_bounds,
                    word_lengths)

                # Infer emphasis scores from word embeddings
                word_embeddings = self.word_decoder(
                    word_embeddings,
                    word_lengths)

            elif emphases.DOWNSAMPLE_LOCATION == 'loss':

                # Downsample activations to word resolution
                word_embeddings = emphases.downsample(
                    frame_embeddings,
                    word_bounds,
                    word_lengths)

            elif emphases.DOWNSAMPLE_LOCATION == 'inference':

                if self.training:

                    # Return frame resolution prominence for framewise loss
                    return self.output_layer(frame_embeddings)

                else:

                    # Downsample activations to word resolution
                    word_embeddings = emphases.downsample(
                        frame_embeddings,
                        word_bounds,
                        word_lengths)

            else:
                raise ValueError(
                    f'Downsample location {emphases.DOWNSAMPLE_LOCATION} ' +
                    'not recognized')

        # Project to scalar
        return self.output_layer(word_embeddings)


###############################################################################
# Utilities
###############################################################################


def mask_from_lengths(lengths):
    """Create boolean mask from sequence lengths"""
    x = torch.arange(lengths.max(), dtype=lengths.dtype, device=lengths.device)
    return (x.unsqueeze(0) < lengths.unsqueeze(1)).unsqueeze(1)



import json
import random
import emphases


###############################################################################
# Partition dataset
###############################################################################


def datasets(datasets):
    """Partition datasets"""
    for dataset in datasets:

        # Check if partition already exists
        file = emphases.PARTITION_DIR / f'{dataset}.json'

        # Random seed
        random.seed(emphases.RANDOM_SEED)

        # Make partition
        if dataset == 'automatic':
            partition = automatic()
        elif dataset == 'buckeye':
            partition = buckeye()
        elif dataset == 'libritts':
            partition = libritts()
        elif dataset == 'crowdsource':
            partition = crowdsource()
        else:
            raise ValueError(f'Dataset {dataset} is not defined')

        # Save to disk
        file.parent.mkdir(exist_ok=True, parents=True)
        with open(file, 'w') as file:
            json.dump(partition, file, ensure_ascii=False, indent=4)


###############################################################################
# Existing datasets
###############################################################################


def buckeye():
    """Partition buckeye dataset"""
    # Get audio files
    directory = emphases.CACHE_DIR / 'buckeye'
    audio_files = directory.rglob('*.wav')

    # Get stems
    stems = [file.stem for file in audio_files]

    # Partition
    return {'train': [], 'valid': [], 'test': stems}


def libritts():
    """Partition libritts dataset"""
    # Get audio files
    directory = emphases.CACHE_DIR / 'libritts'
    audio_files = directory.rglob('*.wav')

    # Get stems
    stems = [file.stem for file in audio_files]

    # Shuffle stems
    random.seed(emphases.RANDOM_SEED)
    random.shuffle(stems)

    # Get split locations
    left = int(emphases.SPLIT_SIZE_TRAIN * len(stems))
    right = left + int(emphases.SPLIT_SIZE_VALID * len(stems))

    # Only train on specified eighth for scaling law experiments
    if emphases.ONE_EIGHTH_UTTERANCES:

        # Partition
        speakers = [str(s) for s in emphases.data.download.LIBRITTS_SPEAKERS]
        train = [stem for stem in stems if stem.split('_')[0] in speakers]
        valid = [stem for stem in stems[left:right] if stem not in train]
        test = [stem for stem in stems[right:] if stem not in train]

    else:

        # Partition
        train = stems[:left]
        valid = stems[left:right]
        test = stems[right:]

    # Maybe limit training set size
    if emphases.MAX_TRAINING_UTTERANCES is not None:
        train = train[:emphases.MAX_TRAINING_UTTERANCES]

    return {'train': train, 'valid': valid, 'test': test}


###############################################################################
# Dataset creation
###############################################################################


def automatic():
    """Partition dataset created from trained model"""
    # Get audio files
    directory = emphases.CACHE_DIR / 'automatic'
    audio_files = directory.rglob('*.wav')

    # Get stems
    stems = [file.stem for file in audio_files]

    # Shuffle stems
    random.seed(emphases.RANDOM_SEED)
    random.shuffle(stems)

    # Get split locations
    left = int(emphases.SPLIT_SIZE_TRAIN * len(stems))
    right = left + int(emphases.SPLIT_SIZE_VALID * len(stems))

    # Partition
    return {
        'train': stems[:left],
        'valid': stems[left:right],
        'test': stems[right:]}


def crowdsource():
    """Partition crowdsourced dataset"""
    # Get audio files
    directory = emphases.CACHE_DIR / 'crowdsource'
    audio_files = directory.rglob('*.wav')

    # Get stems
    stems = [file.stem for file in audio_files]

    # Shuffle stems
    random.seed(emphases.RANDOM_SEED)
    random.shuffle(stems)

    # Get split locations
    left = int(emphases.SPLIT_SIZE_TRAIN * len(stems))
    right = left + int(emphases.SPLIT_SIZE_VALID * len(stems))

    # Only train on specified eighth for scaling law experiments
    if emphases.ONE_EIGHTH_UTTERANCES:

        # Partition
        speakers = [str(s) for s in emphases.data.download.LIBRITTS_SPEAKERS]
        train = [stem for stem in stems if stem.split('_')[0] in speakers]
        valid = [stem for stem in stems[left:right] if stem not in train]
        test = [stem for stem in stems[right:] if stem not in train]

    else:

        # Partition
        train = stems[:left]
        valid = stems[left:right]
        test = stems[right:]

    # Maybe limit training set size
    if emphases.MAX_TRAINING_UTTERANCES is not None:
        train = train[:emphases.MAX_TRAINING_UTTERANCES]

    return {'train': train, 'valid': valid, 'test': test}



import math

import torch

import emphases


###############################################################################
# Transformer stack
###############################################################################


class Transformer(torch.nn.Module):

    def __init__(self, num_layers=emphases.LAYERS, channels=emphases.CHANNELS):
        super().__init__()
        self.position = PositionalEncoding(channels, .1)
        self.model = torch.nn.TransformerEncoder(
            torch.nn.TransformerEncoderLayer(
                channels,
                2,
                dim_feedforward=emphases.CHANNELS),
            num_layers)

    def forward(self, x, lengths):
        mask = emphases.model.mask_from_lengths(lengths)
        return self.model(
            self.position(x.permute(2, 0, 1)),
            src_key_padding_mask=~mask.squeeze(1)
        ).permute(1, 2, 0)


###############################################################################
# Utilities
###############################################################################


class PositionalEncoding(torch.nn.Module):

    def __init__(self, channels, dropout=.1, max_len=5000):
        super().__init__()
        self.dropout = torch.nn.Dropout(p=dropout)
        index = torch.arange(max_len).unsqueeze(1)
        frequency = torch.exp(
            torch.arange(0, channels, 2) * (-math.log(10000.0) / channels))
        encoding = torch.zeros(max_len, 1, channels)
        encoding[:, 0, 0::2] = torch.sin(index * frequency)
        encoding[:, 0, 1::2] = torch.cos(index * frequency)
        self.register_buffer('encoding', encoding)

    def forward(self, x):
        return self.dropout(x + self.encoding[:x.size(0)])
